"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.glusterfs as mod_glusterfs
import salt.states.glusterfs as glusterfs
import salt.utils.cloud
import salt.utils.network
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class GlusterfsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.glusterfs
    """

    def setup_loader_modules(self):
        return {glusterfs: {"__salt__": {"glusterfs.peer": mod_glusterfs.peer}}}

    # 'peered' function tests: 1

    def test_peered(self):
        """
        Test to verify if node is peered.
        """
        name = "server1"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        mock_ip = MagicMock(return_value=["1.2.3.4", "1.2.3.5"])
        mock_ip6 = MagicMock(return_value=["2001:db8::1"])
        mock_host_ips = MagicMock(return_value=["1.2.3.5"])
        mock_peer = MagicMock(return_value=True)
        mock_status = MagicMock(return_value={"uuid1": {"hostnames": [name]}})

        with patch.dict(
            glusterfs.__salt__,
            {"glusterfs.peer_status": mock_status, "glusterfs.peer": mock_peer},
        ):
            with patch.object(salt.utils.network, "ip_addrs", mock_ip), patch.object(
                salt.utils.network, "ip_addrs6", mock_ip6
            ), patch.object(salt.utils.network, "host_to_ips", mock_host_ips):
                comt = "Peering with localhost is not needed"
                ret.update({"comment": comt})
                self.assertDictEqual(glusterfs.peered(name), ret)

                mock_host_ips.return_value = ["127.0.1.1"]
                comt = "Peering with localhost is not needed"
                ret.update({"comment": comt})
                self.assertDictEqual(glusterfs.peered(name), ret)

                mock_host_ips.return_value = ["2001:db8::1"]
                self.assertDictEqual(glusterfs.peered(name), ret)

                mock_host_ips.return_value = ["1.2.3.42"]
                comt = "Host {} already peered".format(name)
                ret.update({"comment": comt})
                self.assertDictEqual(glusterfs.peered(name), ret)

                with patch.dict(glusterfs.__opts__, {"test": False}):
                    old = {"uuid1": {"hostnames": ["other1"]}}
                    new = {
                        "uuid1": {"hostnames": ["other1"]},
                        "uuid2": {"hostnames": ["someAlias", name]},
                    }
                    mock_status.side_effect = [old, new]
                    comt = "Host {} successfully peered".format(name)
                    ret.update({"comment": comt, "changes": {"old": old, "new": new}})
                    self.assertDictEqual(glusterfs.peered(name), ret)
                    mock_status.side_effect = None

                    mock_status.return_value = {"uuid1": {"hostnames": ["other"]}}
                    mock_peer.return_value = False

                    ret.update({"result": False})

                    comt = (
                        "Failed to peer with {0}," + " please check logs for errors"
                    ).format(name)
                    ret.update({"comment": comt, "changes": {}})
                    self.assertDictEqual(glusterfs.peered(name), ret)

                    comt = "Invalid characters in peer name."
                    ret.update({"comment": comt, "name": ":/"})
                    self.assertDictEqual(glusterfs.peered(":/"), ret)
                    ret.update({"name": name})

                with patch.dict(glusterfs.__opts__, {"test": True}):
                    comt = "Peer {} will be added.".format(name)
                    ret.update({"comment": comt, "result": None})
                    self.assertDictEqual(glusterfs.peered(name), ret)

    # 'volume_present' function tests: 1

    def test_volume_present(self):
        """
        Test to ensure that a volume exists
        """
        name = "salt"
        bricks = ["host1:/brick1"]
        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        started_info = {name: {"status": "1"}}
        stopped_info = {name: {"status": "0"}}

        mock_info = MagicMock()
        mock_list = MagicMock()
        mock_create = MagicMock()
        mock_start = MagicMock(return_value=True)

        with patch.dict(
            glusterfs.__salt__,
            {
                "glusterfs.info": mock_info,
                "glusterfs.list_volumes": mock_list,
                "glusterfs.create_volume": mock_create,
                "glusterfs.start_volume": mock_start,
            },
        ):
            with patch.dict(glusterfs.__opts__, {"test": False}):
                mock_list.return_value = [name]
                mock_info.return_value = started_info
                comt = "Volume {} already exists and is started".format(name)
                ret.update({"comment": comt})
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=True), ret
                )

                mock_info.return_value = stopped_info
                comt = "Volume {} already exists and is now started".format(name)
                ret.update(
                    {"comment": comt, "changes": {"old": "stopped", "new": "started"}}
                )
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=True), ret
                )

                comt = "Volume {} already exists".format(name)
                ret.update({"comment": comt, "changes": {}})
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=False), ret
                )
            with patch.dict(glusterfs.__opts__, {"test": True}):
                comt = "Volume {} already exists".format(name)
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=False), ret
                )

                comt = ("Volume {0} already exists" + " and will be started").format(
                    name
                )
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=True), ret
                )

                mock_list.return_value = []
                comt = "Volume {} will be created".format(name)
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=False), ret
                )

                comt = ("Volume {0} will be created" + " and started").format(name)
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=True), ret
                )

            with patch.dict(glusterfs.__opts__, {"test": False}):
                mock_list.side_effect = [[], [name]]
                comt = "Volume {} is created".format(name)
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "changes": {"old": [], "new": [name]},
                    }
                )
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=False), ret
                )

                mock_list.side_effect = [[], [name]]
                comt = "Volume {} is created and is now started".format(name)
                ret.update({"comment": comt, "result": True})
                self.assertDictEqual(
                    glusterfs.volume_present(name, bricks, start=True), ret
                )

                mock_list.side_effect = None
                mock_list.return_value = []
                mock_create.return_value = False
                comt = "Creation of volume {} failed".format(name)
                ret.update({"comment": comt, "result": False, "changes": {}})
                self.assertDictEqual(glusterfs.volume_present(name, bricks), ret)

            with patch.object(
                salt.utils.cloud, "check_name", MagicMock(return_value=True)
            ):
                comt = "Invalid characters in volume name."
                ret.update({"comment": comt, "result": False})
                self.assertDictEqual(glusterfs.volume_present(name, bricks), ret)

    # 'started' function tests: 1

    def test_started(self):
        """
        Test to check if volume has been started
        """
        name = "salt"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        started_info = {name: {"status": "1"}}
        stopped_info = {name: {"status": "0"}}
        mock_info = MagicMock(return_value={})
        mock_start = MagicMock(return_value=True)

        with patch.dict(
            glusterfs.__salt__,
            {"glusterfs.info": mock_info, "glusterfs.start_volume": mock_start},
        ):
            comt = "Volume {} does not exist".format(name)
            ret.update({"comment": comt})
            self.assertDictEqual(glusterfs.started(name), ret)

            mock_info.return_value = started_info
            comt = "Volume {} is already started".format(name)
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(glusterfs.started(name), ret)

            with patch.dict(glusterfs.__opts__, {"test": True}):
                mock_info.return_value = stopped_info
                comt = "Volume {} will be started".format(name)
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(glusterfs.started(name), ret)

            with patch.dict(glusterfs.__opts__, {"test": False}):
                comt = "Volume {} is started".format(name)
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "change": {"new": "started", "old": "stopped"},
                    }
                )
                self.assertDictEqual(glusterfs.started(name), ret)

    # 'add_volume_bricks' function tests: 1

    def test_add_volume_bricks(self):
        """
        Test to add brick(s) to an existing volume
        """
        name = "salt"
        bricks = ["host1:/drive1"]
        old_bricks = ["host1:/drive2"]

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        stopped_volinfo = {"salt": {"status": "0"}}
        volinfo = {
            "salt": {"status": "1", "bricks": {"brick1": {"path": old_bricks[0]}}}
        }
        new_volinfo = {
            "salt": {
                "status": "1",
                "bricks": {
                    "brick1": {"path": old_bricks[0]},
                    "brick2": {"path": bricks[0]},
                },
            }
        }

        mock_info = MagicMock(return_value={})
        mock_add = MagicMock(side_effect=[False, True])

        with patch.dict(
            glusterfs.__salt__,
            {"glusterfs.info": mock_info, "glusterfs.add_volume_bricks": mock_add},
        ):
            ret.update({"comment": "Volume salt does not exist"})
            self.assertDictEqual(glusterfs.add_volume_bricks(name, bricks), ret)

            mock_info.return_value = stopped_volinfo
            ret.update({"comment": "Volume salt is not started"})
            self.assertDictEqual(glusterfs.add_volume_bricks(name, bricks), ret)

            mock_info.return_value = volinfo
            ret.update({"comment": "Adding bricks to volume salt failed"})
            self.assertDictEqual(glusterfs.add_volume_bricks(name, bricks), ret)

            ret.update({"result": True})
            ret.update({"comment": "Bricks already added in volume salt"})
            self.assertDictEqual(glusterfs.add_volume_bricks(name, old_bricks), ret)

            mock_info.side_effect = [volinfo, new_volinfo]
            ret.update(
                {
                    "comment": "Bricks successfully added to volume salt",
                    "changes": {"new": bricks + old_bricks, "old": old_bricks},
                }
            )
            # Let's sort ourselves because the test under python 3 sometimes fails
            # just because of the new changes list order
            result = glusterfs.add_volume_bricks(name, bricks)
            ret["changes"]["new"] = sorted(ret["changes"]["new"])
            result["changes"]["new"] = sorted(result["changes"]["new"])
            self.assertDictEqual(result, ret)

    # 'op_version' function tests: 1

    def test_op_version(self):
        """
        Test setting the Glusterfs op-version
        """
        name = "salt"
        current = 30707
        new = 31200

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        mock_get_version = MagicMock(return_value={})
        mock_set_version = MagicMock(return_value={})

        with patch.dict(
            glusterfs.__salt__,
            {
                "glusterfs.get_op_version": mock_get_version,
                "glusterfs.set_op_version": mock_set_version,
            },
        ):
            mock_get_version.return_value = [False, "some error message"]
            ret.update({"result": False})
            ret.update({"comment": "some error message"})
            self.assertDictEqual(glusterfs.op_version(name, current), ret)

            mock_get_version.return_value = current
            ret.update({"result": True})
            ret.update(
                {
                    "comment": (
                        "Glusterfs cluster.op-version for {} already set to {}".format(
                            name, current
                        )
                    )
                }
            )
            self.assertDictEqual(glusterfs.op_version(name, current), ret)

            with patch.dict(glusterfs.__opts__, {"test": True}):
                mock_set_version.return_value = [False, "Failed to set version"]
                ret.update({"result": None})
                ret.update(
                    {
                        "comment": (
                            "An attempt would be made to set the cluster.op-version for"
                            " {} to {}.".format(name, new)
                        )
                    }
                )
                self.assertDictEqual(glusterfs.op_version(name, new), ret)

            with patch.dict(glusterfs.__opts__, {"test": False}):
                mock_set_version.return_value = [False, "Failed to set version"]
                ret.update({"result": False})
                ret.update({"comment": "Failed to set version"})
                self.assertDictEqual(glusterfs.op_version(name, new), ret)

                mock_set_version.return_value = "some success message"
                ret.update({"comment": "some success message"})
                ret.update({"changes": {"old": current, "new": new}})
                ret.update({"result": True})
                self.assertDictEqual(glusterfs.op_version(name, new), ret)

    # 'max_op_version' function tests: 1

    def test_max_op_version(self):
        """
        Test setting the Glusterfs to its self reported max-op-version
        """
        name = "salt"
        current = 30707
        new = 31200

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        mock_get_version = MagicMock(return_value={})
        mock_get_max_op_version = MagicMock(return_value={})
        mock_set_version = MagicMock(return_value={})

        with patch.dict(
            glusterfs.__salt__,
            {
                "glusterfs.get_op_version": mock_get_version,
                "glusterfs.set_op_version": mock_set_version,
                "glusterfs.get_max_op_version": mock_get_max_op_version,
            },
        ):
            mock_get_version.return_value = [False, "some error message"]
            ret.update({"result": False})
            ret.update({"comment": "some error message"})
            self.assertDictEqual(glusterfs.max_op_version(name), ret)

            mock_get_version.return_value = current
            mock_get_max_op_version.return_value = [False, "some error message"]
            ret.update({"result": False})
            ret.update({"comment": "some error message"})
            self.assertDictEqual(glusterfs.max_op_version(name), ret)

            mock_get_version.return_value = current
            mock_get_max_op_version.return_value = current
            ret.update({"result": True})
            ret.update(
                {
                    "comment": (
                        "The cluster.op-version is already set to the"
                        " cluster.max-op-version of {}".format(current)
                    )
                }
            )
            self.assertDictEqual(glusterfs.max_op_version(name), ret)

            with patch.dict(glusterfs.__opts__, {"test": True}):
                mock_get_max_op_version.return_value = new
                ret.update({"result": None})
                ret.update(
                    {
                        "comment": (
                            "An attempt would be made to set the cluster.op-version"
                            " to {}.".format(new)
                        )
                    }
                )
                self.assertDictEqual(glusterfs.max_op_version(name), ret)

            with patch.dict(glusterfs.__opts__, {"test": False}):
                mock_set_version.return_value = [False, "Failed to set version"]
                ret.update({"result": False})
                ret.update({"comment": "Failed to set version"})
                self.assertDictEqual(glusterfs.max_op_version(name), ret)

                mock_set_version.return_value = "some success message"
                ret.update({"comment": "some success message"})
                ret.update({"changes": {"old": current, "new": new}})
                ret.update({"result": True})
                self.assertDictEqual(glusterfs.max_op_version(name), ret)



"""
An engine that continuously reads messages from SQS and fires them as events.

Note that long polling is utilized to avoid excessive CPU usage.

.. versionadded:: 2015.8.0

:depends: boto

Configuration
=============

This engine can be run on the master or on a minion.

Example Config:

.. code-block:: yaml

    sqs.keyid: GKTADJGHEIQSXMKKRBJ08H
    sqs.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
    sqs.message_format: json

Explicit sqs credentials are accepted but this engine can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More Information available at::

   http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

If IAM roles are not (or for ``boto`` version < 2.5.1) used you need to
specify them either in a pillar or in the config file of the master or
minion, as appropriate:

To deserialize the message from json:

.. code-block:: yaml

    sqs.message_format: json

It's also possible to specify key, keyid and region via a profile:

.. code-block:: yaml

    sqs.keyid: GKTADJGHEIQSXMKKRBJ08H
    sqs.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

A region may also be specified in the configuration:

.. code-block:: yaml

    sqs.region: us-east-1

If a region is not specified, the default is us-east-1.

It's also possible to specify key, keyid and region via a profile:

.. code-block:: yaml

    myprofile:
      keyid: GKTADJGHEIQSXMKKRBJ08H
      key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
      region: us-east-1

Additionally you can define cross account sqs:

.. code-block:: yaml

    engines:
      - sqs_events:
          queue: prod
          owner_acct_id: 111111111111

"""

import logging
import time

import salt.utils.event
import salt.utils.json

try:
    import boto.sqs

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False


def __virtual__():
    if not HAS_BOTO:
        return (
            False,
            "Cannot import engine sqs_events because the required boto module is"
            " missing",
        )
    else:
        return True


log = logging.getLogger(__name__)


def _get_sqs_conn(profile, region=None, key=None, keyid=None):
    """
    Get a boto connection to SQS.
    """
    if profile:
        if isinstance(profile, str):
            _profile = __opts__[profile]
        elif isinstance(profile, dict):
            _profile = profile
        key = _profile.get("key", None)
        keyid = _profile.get("keyid", None)
        region = _profile.get("region", None)

    if not region:
        region = __opts__.get("sqs.region", "us-east-1")
    if not key:
        key = __opts__.get("sqs.key", None)
    if not keyid:
        keyid = __opts__.get("sqs.keyid", None)
    try:
        conn = boto.sqs.connect_to_region(
            region, aws_access_key_id=keyid, aws_secret_access_key=key
        )
    except boto.exception.NoAuthHandlerFound:
        log.error(
            "No authentication credentials found when attempting to"
            " make sqs_event engine connection to AWS."
        )
        return None
    return conn


def _process_queue(
    q,
    q_name,
    fire_master,
    tag="salt/engine/sqs",
    owner_acct_id=None,
    message_format=None,
):
    if not q:
        log.warning(
            "failure connecting to queue: %s, waiting 10 seconds.",
            ":".join([_f for _f in (str(owner_acct_id), q_name) if _f]),
        )
        time.sleep(10)
    else:
        msgs = q.get_messages(wait_time_seconds=20)
        for msg in msgs:
            if message_format == "json":
                fire_master(
                    tag=tag, data={"message": salt.utils.json.loads(msg.get_body())}
                )
            else:
                fire_master(tag=tag, data={"message": msg.get_body()})
            msg.delete()


def start(queue, profile=None, tag="salt/engine/sqs", owner_acct_id=None):
    """
    Listen to sqs and fire message on event bus
    """
    if __opts__.get("__role") == "master":
        fire_master = salt.utils.event.get_master_event(
            __opts__, __opts__["sock_dir"], listen=False
        ).fire_event
    else:
        fire_master = __salt__["event.send"]

    message_format = __opts__.get("sqs.message_format", None)

    sqs = _get_sqs_conn(profile)
    q = None
    while True:
        if not q:
            q = sqs.get_queue(queue, owner_acct_id=owner_acct_id)
            q.set_message_class(boto.sqs.message.RawMessage)

        _process_queue(
            q,
            queue,
            fire_master,
            tag=tag,
            owner_acct_id=owner_acct_id,
            message_format=message_format,
        )



import pytest

import salt.utils.win_dacl as win_dacl
from tests.support.mock import patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        win_dacl: {
            "__opts__": minion_opts,
        },
    }


@pytest.fixture(scope="function")
def test_file():
    with pytest.helpers.temp_file("dacl_test.file") as test_file:
        yield test_file


@pytest.fixture(scope="function")
def test_dir(tmp_path_factory):
    test_dir = tmp_path_factory.mktemp("test_dir")
    yield str(test_dir)
    test_dir.rmdir()


def test_get_set_owner(test_file):
    result = win_dacl.set_owner(
        obj_name=str(test_file), principal="Backup Operators", obj_type="file"
    )
    assert result is True
    result = win_dacl.get_owner(obj_name=str(test_file), obj_type="file")
    assert result == "Backup Operators"


def test_get_set_primary_group(test_file):
    result = win_dacl.set_primary_group(
        obj_name=str(test_file), principal="Backup Operators", obj_type="file"
    )
    assert result is True
    result = win_dacl.get_primary_group(obj_name=str(test_file), obj_type="file")
    assert result == "Backup Operators"


def test_get_set_permissions(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "This folder only",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_applies_to_this_folder_only(test_dir):
    """
    #60103
    """
    result = win_dacl.set_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        applies_to="this_folder_only",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "This folder only",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_applies_to_this_folder_files(test_dir):
    """
    #60103
    """
    result = win_dacl.set_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        applies_to="this_folder_files",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "This folder and files",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_applies_to_this_folder_subfolders(test_dir):
    """
    #60103
    """
    result = win_dacl.set_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        applies_to="this_folder_subfolders",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "This folder and subfolders",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_applies_to_this_folder_subfolders_files(test_dir):
    """
    #60103
    """
    result = win_dacl.set_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        applies_to="this_folder_subfolders_files",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_applies_to_files_only(test_dir):
    """
    #60103
    """
    result = win_dacl.set_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        applies_to="files_only",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "Files only",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_applies_to_subfolders_only(test_dir):
    """
    #60103
    """
    result = win_dacl.set_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        applies_to="subfolders_only",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "Subfolders only",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_applies_to_subfolders_files(test_dir):
    """
    #60103
    """
    result = win_dacl.set_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        applies_to="subfolders_files",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "Subfolders and files only",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=test_dir,
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected


def test_has_permission_exact_match(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permission exact
    result = win_dacl.has_permission(
        obj_name=str(test_file),
        principal="Backup Operators",
        permission="full_control",
        access_mode="grant",
        obj_type="file",
        exact=True,
    )
    assert result is True


def test_has_permission_exact_no_match(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permission exact
    result = win_dacl.has_permission(
        obj_name=str(test_file),
        principal="Backup Operators",
        permission="read",
        access_mode="grant",
        obj_type="file",
        exact=True,
    )
    assert result is False


def test_has_permission_contains(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permission exact
    result = win_dacl.has_permission(
        obj_name=str(test_file),
        principal="Backup Operators",
        permission="read",
        access_mode="grant",
        obj_type="file",
        exact=False,
    )
    assert result is True


def test_has_permission_contains_advanced(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permission exact
    result = win_dacl.has_permission(
        obj_name=str(test_file),
        principal="Backup Operators",
        permission="read_data",
        access_mode="grant",
        obj_type="file",
        exact=False,
    )
    assert result is True


def test_has_permission_missing(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="read_execute",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permission not exact
    result = win_dacl.has_permission(
        obj_name=str(test_file),
        principal="Backup Operators",
        permission="write",
        access_mode="grant",
        obj_type="file",
        exact=False,
    )
    assert result is False


def test_has_permission_missing_advanced(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="read_execute",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permission not exact
    result = win_dacl.has_permission(
        obj_name=str(test_file),
        principal="Backup Operators",
        permission="write_data",
        access_mode="grant",
        obj_type="file",
        exact=False,
    )


def test_has_permissions_contains(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permissions exact
    result = win_dacl.has_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read", "write"],
        access_mode="grant",
        obj_type="file",
        exact=False,
    )
    assert result is True


def test_has_permissions_contains_advanced(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permissions exact
    result = win_dacl.has_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data"],
        access_mode="grant",
        obj_type="file",
        exact=False,
    )
    assert result is True


def test_has_permissions_missing(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="read_execute",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permissions exact
    result = win_dacl.has_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data"],
        access_mode="grant",
        obj_type="file",
        exact=False,
    )
    assert result is False


def test_has_permissions_exact_contains(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data"],
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permissions exact
    result = win_dacl.has_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data"],
        access_mode="grant",
        obj_type="file",
        exact=True,
    )
    assert result is True


def test_has_permissions_exact_has_extra(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data", "create_folders"],
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permissions exact
    result = win_dacl.has_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data"],
        access_mode="grant",
        obj_type="file",
        exact=True,
    )
    assert result is False


def test_has_permissions_exact_missing(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data"],
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    # Test has_permissions exact
    result = win_dacl.has_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions=["read_data", "write_data", "create_folders"],
        access_mode="grant",
        obj_type="file",
        exact=True,
    )
    assert result is False


def test_rm_permissions(test_file):
    result = win_dacl.set_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        permissions="full_control",
        access_mode="grant",
        obj_type="file",
        reset_perms=False,
        protected=None,
    )
    assert result is True

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "grant": {
                    "applies to": "This folder only",
                    "permissions": "Full control",
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected

    result = win_dacl.rm_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        obj_type="file",
    )
    assert result is True
    result = win_dacl.get_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == {}


def test_get_set_inheritance(test_file):
    result = win_dacl.set_inheritance(
        obj_name=str(test_file),
        enabled=True,
        obj_type="file",
        clear=False,
    )
    assert result is True

    result = win_dacl.get_inheritance(obj_name=str(test_file), obj_type="file")
    assert result is True

    result = win_dacl.set_inheritance(
        obj_name=str(test_file),
        enabled=False,
        obj_type="file",
        clear=False,
    )
    assert result is True

    result = win_dacl.get_inheritance(obj_name=str(test_file), obj_type="file")
    assert result is False


def test_copy_security():
    with pytest.helpers.temp_file("source_test.file") as source:
        # Set permissions on Source
        result = win_dacl.set_permissions(
            obj_name=str(source),
            principal="Backup Operators",
            permissions="full_control",
            access_mode="grant",
            obj_type="file",
            reset_perms=False,
            protected=None,
        )
        assert result is True

        # Set owner on Source
        result = win_dacl.set_owner(
            obj_name=str(source), principal="Backup Operators", obj_type="file"
        )
        assert result is True

        # Set group on Source
        result = win_dacl.set_primary_group(
            obj_name=str(source), principal="Backup Operators", obj_type="file"
        )
        assert result is True

        with pytest.helpers.temp_file("target_test.file") as target:
            # Copy security from Source to Target
            result = win_dacl.copy_security(source=str(source), target=str(target))
            assert result is True

            # Verify permissions on Target
            expected = {
                "Not Inherited": {
                    "Backup Operators": {
                        "grant": {
                            "applies to": "This folder only",
                            "permissions": "Full control",
                        }
                    }
                }
            }
            result = win_dacl.get_permissions(
                obj_name=str(target),
                principal="Backup Operators",
                obj_type="file",
            )
            assert result == expected

            # Verify owner on Target
            result = win_dacl.get_owner(obj_name=str(target), obj_type="file")
            assert result == "Backup Operators"

            # Verify group on Target
            result = win_dacl.get_primary_group(obj_name=str(target), obj_type="file")
            assert result == "Backup Operators"


def test_check_perms(test_file):
    result = win_dacl.check_perms(
        obj_name=str(test_file),
        obj_type="file",
        ret={},
        owner="Users",
        grant_perms={"Backup Operators": {"perms": "read"}},
        deny_perms={
            "Backup Operators": {"perms": ["delete"]},
            "NETWORK SERVICE": {
                "perms": [
                    "delete",
                    "change_permissions",
                    "write_attributes",
                    "write_data",
                ]
            },
        },
        inheritance=True,
        reset=False,
    )

    expected = {
        "changes": {
            "owner": "Users",
            "grant_perms": {"Backup Operators": {"permissions": "read"}},
            "deny_perms": {
                "Backup Operators": {"permissions": ["delete"]},
                "NETWORK SERVICE": {
                    "permissions": [
                        "delete",
                        "change_permissions",
                        "write_attributes",
                        "write_data",
                    ]
                },
            },
        },
        "comment": "",
        "name": str(test_file),
        "result": True,
    }
    assert result == expected

    expected = {
        "Not Inherited": {
            "Backup Operators": {
                "deny": {
                    "applies to": "This folder only",
                    "permissions": ["Delete"],
                },
                "grant": {"applies to": "This folder only", "permissions": "Read"},
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == expected

    expected = {
        "Not Inherited": {
            "NETWORK SERVICE": {
                "deny": {
                    "applies to": "This folder only",
                    "permissions": [
                        "Change permissions",
                        "Create files / write data",
                        "Delete",
                        "Write attributes",
                    ],
                }
            }
        }
    }
    result = win_dacl.get_permissions(
        obj_name=str(test_file),
        principal="NETWORK SERVICE",
        obj_type="file",
    )
    assert result == expected

    result = win_dacl.get_owner(obj_name=str(test_file), obj_type="file")
    assert result == "Users"


def test_check_perms_test_true(test_file):
    with patch.dict(win_dacl.__opts__, {"test": True}):
        result = win_dacl.check_perms(
            obj_name=str(test_file),
            obj_type="file",
            ret=None,
            owner="Users",
            grant_perms={"Backup Operators": {"perms": "read"}},
            deny_perms={
                "NETWORK SERVICE": {
                    "perms": ["delete", "set_value", "write_dac", "write_owner"]
                },
                "Backup Operators": {"perms": ["delete"]},
            },
            inheritance=True,
            reset=False,
        )

    expected = {
        "changes": {
            "owner": "Users",
            "grant_perms": {"Backup Operators": {"permissions": "read"}},
            "deny_perms": {
                "Backup Operators": {"permissions": ["delete"]},
                "NETWORK SERVICE": {
                    "permissions": [
                        "delete",
                        "set_value",
                        "write_dac",
                        "write_owner",
                    ]
                },
            },
        },
        "comment": "",
        "name": str(test_file),
        "result": None,
    }
    assert result == expected

    result = win_dacl.get_owner(obj_name=str(test_file), obj_type="file")
    assert result != "Users"

    result = win_dacl.get_permissions(
        obj_name=str(test_file),
        principal="Backup Operators",
        obj_type="file",
    )
    assert result == {}


def test_set_perms(test_file):
    result = win_dacl.set_perms(
        obj_name=str(test_file),
        obj_type="file",
        grant_perms={"Backup Operators": {"perms": "read"}},
        deny_perms={
            "NETWORK SERVICE": {
                "perms": [
                    "delete",
                    "change_permissions",
                    "write_attributes",
                    "write_data",
                ]
            }
        },
        inheritance=True,
        reset=False,
    )

    expected = {
        "deny": {
            "NETWORK SERVICE": {
                "perms": [
                    "delete",
                    "change_permissions",
                    "write_attributes",
                    "write_data",
                ]
            }
        },
        "grant": {"Backup Operators": {"perms": "read"}},
    }
    assert result == expected



"""
Test the scan roster.
"""

import socket

import pytest

import salt.roster.scan as scan_
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {scan_: {"__opts__": {"ssh_scan_ports": "22", "ssh_scan_timeout": 0.01}}}


def test_single_ip():
    """Test that minion files in the directory roster match and render."""
    with patch("salt.utils.network.get_socket"):
        ret = scan_.targets("127.0.0.1")
    assert ret == {"127.0.0.1": {"host": "127.0.0.1", "port": 22}}


def test_single_network():
    """Test that minion files in the directory roster match and render."""
    with patch("salt.utils.network.get_socket"):
        ret = scan_.targets("127.0.0.0/30")
    assert ret == {
        "127.0.0.1": {"host": "127.0.0.1", "port": 22},
        "127.0.0.2": {"host": "127.0.0.2", "port": 22},
    }


def test_multiple_ips():
    """Test that minion files in the directory roster match and render."""
    with patch("salt.utils.network.get_socket"):
        ret = scan_.targets(["127.0.0.1", "127.0.0.2"], tgt_type="list")
    assert ret == {
        "127.0.0.1": {"host": "127.0.0.1", "port": 22},
        "127.0.0.2": {"host": "127.0.0.2", "port": 22},
    }


def test_multiple_networks():
    """Test that minion files in the directory roster match and render."""
    with patch("salt.utils.network.get_socket"):
        ret = scan_.targets(
            ["127.0.0.0/30", "127.0.2.1", "127.0.1.0/30"], tgt_type="list"
        )
    assert ret == {
        "127.0.0.1": {"host": "127.0.0.1", "port": 22},
        "127.0.0.2": {"host": "127.0.0.2", "port": 22},
        "127.0.2.1": {"host": "127.0.2.1", "port": 22},
        "127.0.1.1": {"host": "127.0.1.1", "port": 22},
        "127.0.1.2": {"host": "127.0.1.2", "port": 22},
    }


def test_malformed_ip():
    """Test that minion files in the directory roster match and render."""
    with patch("salt.utils.network.get_socket"):
        ret = scan_.targets("127001")
    assert ret == {}


def test_multiple_with_malformed():
    """Test that minion files in the directory roster match and render."""
    with patch("salt.utils.network.get_socket"):
        ret = scan_.targets(["127.0.0.1", "127002", "127.0.1.0/30"], tgt_type="list")
    assert ret == {
        "127.0.0.1": {"host": "127.0.0.1", "port": 22},
        "127.0.1.1": {"host": "127.0.1.1", "port": 22},
        "127.0.1.2": {"host": "127.0.1.2", "port": 22},
    }


def test_multiple_no_connection():
    """Test that minion files in the directory roster match and render."""
    socket_mock = MagicMock()
    socket_mock.connect = MagicMock(
        side_effect=[None, socket.error(), None, socket.error(), None]
    )
    with patch("salt.utils.network.get_socket", return_value=socket_mock):
        ret = scan_.targets(
            ["127.0.0.0/30", "127.0.2.1", "127.0.1.0/30"], tgt_type="list"
        )
    assert ret == {
        "127.0.0.1": {"host": "127.0.0.1", "port": 22},
        "127.0.0.2": {},
        "127.0.2.1": {"host": "127.0.2.1", "port": 22},
        "127.0.1.1": {},
        "127.0.1.2": {"host": "127.0.1.2", "port": 22},
    }



import os
from tempfile import NamedTemporaryFile

import pytest

import salt.modules.timezone as timezone
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, mock_open, patch

GET_ZONE_FILE = "salt.modules.timezone._get_zone_file"
GET_LOCALTIME_PATH = "salt.modules.timezone._get_localtime_path"
TEST_TZ = "UTC"


@pytest.fixture
def configure_loader_modules():
    return {
        timezone: {
            "__grains__": {"os": "", "os_family": "Ubuntu"},
            "__salt__": {
                "file.sed": MagicMock(),
                "cmd.run": MagicMock(),
                "cmd.retcode": MagicMock(return_value=0),
            },
        }
    }


@pytest.fixture
def tempfiles():
    tempfiles = []
    yield tempfiles
    for tempfile in tempfiles:
        try:
            os.remove(tempfile.name)
        except OSError:
            pass
    del tempfiles


@pytest.fixture
def patch_os():
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        with patch("os.path.exists", MagicMock(return_value=True)):
            with patch("os.unlink", MagicMock()):
                with patch("os.symlink", MagicMock()):
                    yield


def test_zone_compare_equal(tempfiles):
    etc_localtime = create_tempfile_with_contents("a", tempfiles)
    zone_path = create_tempfile_with_contents("a", tempfiles)

    with patch(GET_ZONE_FILE, lambda p: zone_path.name):
        with patch(GET_LOCALTIME_PATH, lambda: etc_localtime.name):

            assert timezone.zone_compare("foo")


def test_zone_compare_arch(tempfiles):
    """
    test zone_compare function when OS is arch
    """
    etc_localtime = create_tempfile_with_contents("a", tempfiles)
    zone_path = create_tempfile_with_contents("a", tempfiles)
    mock_zone = MagicMock(return_value="foo")
    patch_zone = patch.object(timezone, "get_zone", mock_zone)

    with patch_zone:
        with patch.dict(timezone.__grains__, {"os_family": "Arch"}):
            with patch(GET_ZONE_FILE, lambda p: zone_path.name):
                with patch(GET_LOCALTIME_PATH, lambda: etc_localtime.name):
                    with patch("os.path.isfile", return_value=False):
                        assert timezone.zone_compare("foo")
                        mock_zone.assert_called()


def test_zone_compare_nonexistent(tempfiles):
    etc_localtime = create_tempfile_with_contents("a", tempfiles)

    with patch(GET_ZONE_FILE, lambda p: "/foopath/nonexistent"):
        with patch(GET_LOCALTIME_PATH, lambda: etc_localtime.name):

            pytest.raises(SaltInvocationError, timezone.zone_compare, "foo")


def test_zone_compare_unequal(tempfiles):
    etc_localtime = create_tempfile_with_contents("a", tempfiles)
    zone_path = create_tempfile_with_contents("b", tempfiles)

    with patch(GET_ZONE_FILE, lambda p: zone_path.name):
        with patch(GET_LOCALTIME_PATH, lambda: etc_localtime.name):

            assert not timezone.zone_compare("foo")


def test_missing_localtime():
    with patch(GET_ZONE_FILE, lambda p: "/nonexisting"):
        with patch(GET_LOCALTIME_PATH, lambda: "/also-missing"):
            pytest.raises(CommandExecutionError, timezone.zone_compare, "foo")


def create_tempfile_with_contents(contents, tempfiles=None):
    temp = NamedTemporaryFile(delete=False)
    temp.write(salt.utils.stringutils.to_bytes(contents))
    temp.close()
    tempfiles.append(temp)
    return temp


def test_get_zone_centos():
    """
    Test CentOS is recognized
    :return:
    """
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        with patch.dict(timezone.__grains__, {"os": "centos"}):
            with patch(
                "salt.modules.timezone._get_zone_etc_localtime",
                MagicMock(return_value=TEST_TZ),
            ):
                assert timezone.get_zone() == TEST_TZ


def test_get_zone_os_family_rh_suse():
    """
    Test RedHat and Suse are recognized
    :return:
    """
    for osfamily in ["RedHat", "Suse"]:
        with patch("salt.utils.path.which", MagicMock(return_value=False)):
            with patch.dict(timezone.__grains__, {"os_family": [osfamily]}):
                with patch(
                    "salt.modules.timezone._get_zone_sysconfig",
                    MagicMock(return_value=TEST_TZ),
                ):
                    assert timezone.get_zone() == TEST_TZ


def test_get_zone_os_family_debian_gentoo():
    """
    Test Debian and Gentoo are recognized
    :return:
    """
    for osfamily in ["Debian", "Gentoo"]:
        with patch("salt.utils.path.which", MagicMock(return_value=False)):
            with patch.dict(timezone.__grains__, {"os_family": [osfamily]}):
                with patch(
                    "salt.modules.timezone._get_zone_etc_timezone",
                    MagicMock(return_value=TEST_TZ),
                ):
                    assert timezone.get_zone() == TEST_TZ


def test_get_zone_os_family_allbsd_nilinuxrt_slackware():
    """
    Test *BSD, NILinuxRT and Slackware are recognized
    :return:
    """
    for osfamily in ["FreeBSD", "OpenBSD", "NetBSD", "NILinuxRT", "Slackware"]:
        with patch("salt.utils.path.which", MagicMock(return_value=False)):
            with patch.dict(timezone.__grains__, {"os_family": osfamily}):
                with patch(
                    "salt.modules.timezone._get_zone_etc_localtime",
                    MagicMock(return_value=TEST_TZ),
                ):
                    assert timezone.get_zone() == TEST_TZ


def test_get_zone_os_family_slowlaris():
    """
    Test Slowlaris is recognized
    :return:
    """
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        with patch.dict(timezone.__grains__, {"os_family": ["Solaris"]}):
            with patch(
                "salt.modules.timezone._get_zone_solaris",
                MagicMock(return_value=TEST_TZ),
            ):
                assert timezone.get_zone() == TEST_TZ


def test_get_zone_os_family_aix():
    """
    Test IBM AIX is recognized
    :return:
    """
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        with patch.dict(timezone.__grains__, {"os_family": ["AIX"]}):
            with patch(
                "salt.modules.timezone._get_zone_aix",
                MagicMock(return_value=TEST_TZ),
            ):
                assert timezone.get_zone() == TEST_TZ


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_zone_os_family_nilinuxrt(patch_os):
    """
    Test zone set on NILinuxRT
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["NILinuxRT"]}), patch.dict(
        timezone.__grains__, {"lsb_distrib_id": "nilrt"}
    ):
        assert timezone.set_zone(TEST_TZ)


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_zone_os_family_allbsd_slackware(patch_os):
    """
    Test zone set on *BSD and Slackware
    :return:
    """
    for osfamily in ["FreeBSD", "OpenBSD", "NetBSD", "Slackware"]:
        with patch.dict(timezone.__grains__, {"os_family": osfamily}):
            assert timezone.set_zone(TEST_TZ)


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_zone_redhat(patch_os):
    """
    Test zone set on RH series
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["RedHat"]}):
        assert timezone.set_zone(TEST_TZ)
        name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
        assert args == ("/etc/sysconfig/clock", "^ZONE=.*", 'ZONE="UTC"')


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_zone_suse(patch_os):
    """
    Test zone set on SUSE series
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["Suse"]}):
        assert timezone.set_zone(TEST_TZ)
        name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
        assert args == ("/etc/sysconfig/clock", "^TIMEZONE=.*", 'TIMEZONE="UTC"')


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_zone_gentoo(patch_os):
    """
    Test zone set on Gentoo series
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["Gentoo"]}):
        with patch("salt.utils.files.fopen", mock_open()) as m_open:
            assert timezone.set_zone(TEST_TZ)
            fh_ = m_open.filehandles["/etc/timezone"][0]
            assert fh_.call.args == ("/etc/timezone", "w"), fh_.call.args
            assert fh_.write_calls == ["UTC", "\n"], fh_.write_calls


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_zone_debian(patch_os):
    """
    Test zone set on Debian series
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["Debian"]}):
        with patch("salt.utils.files.fopen", mock_open()) as m_open:
            assert timezone.set_zone(TEST_TZ)
            fh_ = m_open.filehandles["/etc/timezone"][0]
            assert fh_.call.args == ("/etc/timezone", "w"), fh_.call.args
            assert fh_.write_calls == ["UTC", "\n"], fh_.write_calls


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_get_hwclock_timedate_utc():
    """
    Test get hwclock UTC/localtime
    :return:
    """
    with patch("salt.utils.path.which", MagicMock(return_value=True)):
        with patch("os.path.exists", MagicMock(return_value=True)):
            with patch("os.unlink", MagicMock()):
                with patch("os.symlink", MagicMock()):
                    with patch(
                        "salt.modules.timezone._timedatectl",
                        MagicMock(return_value={"stdout": "rtc in local tz"}),
                    ):
                        assert timezone.get_hwclock() == "UTC"
                    with patch(
                        "salt.modules.timezone._timedatectl",
                        MagicMock(return_value={"stdout": "rtc in local tz:yes"}),
                    ):
                        assert timezone.get_hwclock() == "localtime"


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_get_hwclock_suse(patch_os):
    """
    Test get hwclock on SUSE
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["Suse"]}):
        timezone.get_hwclock()
        name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[0]
        assert args == (["tail", "-n", "1", "/etc/adjtime"],)
        assert kwarg == {"python_shell": False}


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_get_hwclock_redhat(patch_os):
    """
    Test get hwclock on RedHat
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["RedHat"]}):
        timezone.get_hwclock()
        name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[0]
        assert args == (["tail", "-n", "1", "/etc/adjtime"],)
        assert kwarg == {"python_shell": False}


def _test_get_hwclock_debian(
    patch_os,
):  # TODO: Enable this when testing environment is working properly
    """
    Test get hwclock on Debian
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["Debian"]}):
        timezone.get_hwclock()
        name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[0]
        assert args == (["tail", "-n", "1", "/etc/adjtime"],)
        assert kwarg == {"python_shell": False}


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_get_hwclock_solaris(patch_os):
    """
    Test get hwclock on Solaris
    :return:
    """
    # Incomplete
    with patch.dict(timezone.__grains__, {"os_family": ["Solaris"]}):
        assert timezone.get_hwclock() == "UTC"
        with patch("salt.utils.files.fopen", mock_open()):
            assert timezone.get_hwclock() == "localtime"


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_get_hwclock_aix(patch_os):
    """
    Test get hwclock on AIX
    :return:
    """
    # Incomplete
    hwclock = "localtime"
    if not os.path.isfile("/etc/environment"):
        hwclock = "UTC"
    with patch.dict(timezone.__grains__, {"os_family": ["AIX"]}):
        assert timezone.get_hwclock() == hwclock


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_get_hwclock_slackware_with_adjtime(patch_os):
    """
    Test get hwclock on Slackware with /etc/adjtime present
    :return:
    """
    with patch.dict(timezone.__grains__, {"os_family": ["Slackware"]}):
        timezone.get_hwclock()
        name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[0]
        assert args == (["tail", "-n", "1", "/etc/adjtime"],)
        assert kwarg == {"python_shell": False}


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_get_hwclock_slackware_without_adjtime():
    """
    Test get hwclock on Slackware without /etc/adjtime present
    :return:
    """
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        with patch("os.path.exists", MagicMock(return_value=False)):
            with patch("os.unlink", MagicMock()):
                with patch("os.symlink", MagicMock()):
                    with patch.dict(timezone.__grains__, {"os_family": ["Slackware"]}):
                        with patch(
                            "salt.utils.files.fopen", mock_open(read_data="UTC")
                        ):
                            assert timezone.get_hwclock() == "UTC"
                        with patch(
                            "salt.utils.files.fopen", mock_open(read_data="localtime")
                        ):
                            assert timezone.get_hwclock() == "localtime"


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_timedatectl():
    """
    Test set hwclock with timedatectl
    :return:
    """
    with patch("salt.utils.path.which", MagicMock(return_value=True)):
        timezone.set_hwclock("UTC")
        name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[0]
    assert args == (["timedatectl", "set-local-rtc", "false"],)

    with patch("salt.utils.path.which", MagicMock(return_value=True)):
        timezone.set_hwclock("localtime")
    with patch("salt.utils.path.which", MagicMock(return_value=True)):
        name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[1]
    assert args == (["timedatectl", "set-local-rtc", "true"],)


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_aix_nilinuxrt(patch_os):
    """
    Test set hwclock on AIX and NILinuxRT
    :return:
    """
    for osfamily in ["AIX", "NILinuxRT"]:
        with patch.dict(timezone.__grains__, {"os_family": osfamily}):
            with pytest.raises(SaltInvocationError):
                assert timezone.set_hwclock("forty two")
            assert timezone.set_hwclock("UTC")


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_solaris(patch_os):
    """
    Test set hwclock on Solaris
    :return:
    """
    with patch(
        "salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE")
    ):
        with patch.dict(
            timezone.__grains__, {"os_family": ["Solaris"], "cpuarch": "x86"}
        ):
            with pytest.raises(SaltInvocationError):
                assert timezone.set_hwclock("forty two")
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[0]
            assert args == (["rtc", "-z", "GMT"],)
            assert kwargs == {"python_shell": False}


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_arch(patch_os):
    """
    Test set hwclock on arch
    :return:
    """
    with patch(
        "salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE")
    ):
        with patch.dict(timezone.__grains__, {"os_family": ["Arch"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[0]
            assert args == (["timezonectl", "set-local-rtc", "false"],)
            assert kwargs == {"python_shell": False}


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_redhat(patch_os):
    """
    Test set hwclock on RedHat
    :return:
    """
    with patch(
        "salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE")
    ):
        with patch.dict(timezone.__grains__, {"os_family": ["RedHat"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/sysconfig/clock", "^ZONE=.*", 'ZONE="TEST_TIMEZONE"')


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_suse(patch_os):
    """
    Test set hwclock on SUSE
    :return:
    """
    with patch(
        "salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE")
    ):
        with patch.dict(timezone.__grains__, {"os_family": ["Suse"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == (
                "/etc/sysconfig/clock",
                "^TIMEZONE=.*",
                'TIMEZONE="TEST_TIMEZONE"',
            )


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_debian(patch_os):
    """
    Test set hwclock on Debian
    :return:
    """
    with patch(
        "salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE")
    ):
        with patch.dict(timezone.__grains__, {"os_family": ["Debian"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/default/rcS", "^UTC=.*", "UTC=yes")

            assert timezone.set_hwclock("localtime")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[1]
            assert args == ("/etc/default/rcS", "^UTC=.*", "UTC=no")


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_gentoo(patch_os):
    """
    Test set hwclock on Gentoo
    :return:
    """
    with patch(
        "salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE")
    ):
        with patch.dict(timezone.__grains__, {"os_family": ["Gentoo"]}):
            with pytest.raises(SaltInvocationError):
                timezone.set_hwclock("forty two")

            timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/conf.d/hwclock", "^clock=.*", 'clock="UTC"')

            timezone.set_hwclock("localtime")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[1]
            assert args == ("/etc/conf.d/hwclock", "^clock=.*", 'clock="local"')


@pytest.mark.skip_on_windows(reason="os.symlink not available in Windows")
def test_set_hwclock_slackware(patch_os):
    """
    Test set hwclock on Slackware
    :return:
    """
    with patch(
        "salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE")
    ):
        with patch.dict(timezone.__grains__, {"os_family": ["Slackware"]}):
            with pytest.raises(SaltInvocationError):
                timezone.set_hwclock("forty two")

            timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/hardwareclock", "^(UTC|localtime)", "UTC")

            timezone.set_hwclock("localtime")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[1]
            assert args == ("/etc/hardwareclock", "^(UTC|localtime)", "localtime")



"""
Boto3 Common Utils
=================

Note: This module depends on the dicts packed by the loader and,
therefore, must be accessed via the loader or from the __utils__ dict.

This module provides common functionality for the boto execution modules.
The expected usage is to call `apply_funcs` from the `__virtual__` function
of the module. This will bring properly initilized partials of  `_get_conn`
and `_cache_id` into the module's namespace.

Example Usage:

    .. code-block:: python

        def __virtual__():
            __utils__['boto.apply_funcs'](__name__, 'vpc')

        def test():
            conn = _get_conn()
            vpc_id = _cache_id('test-vpc')

.. versionadded:: 2015.8.0
"""


import hashlib
import logging
import sys
from functools import partial

import salt.loader.context
import salt.utils.stringutils
import salt.utils.versions
from salt.exceptions import SaltInvocationError

# pylint: disable=import-error
try:
    # pylint: disable=import-error
    import boto3
    import boto3.session
    import botocore  # pylint: disable=W0611
    import botocore.exceptions

    # pylint: enable=import-error
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False
# pylint: enable=import-error


log = logging.getLogger(__name__)

__virtualname__ = "boto3"
__salt_loader__ = salt.loader.context.LoaderContext()
__context__ = __salt_loader__.named_context("__context__", {})


def __virtual__():
    """
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    """
    has_boto = salt.utils.versions.check_boto_reqs()
    if has_boto is True:
        return __virtualname__
    return has_boto


def _option(value):
    """
    Look up the value for an option.
    """
    if value in __opts__:
        return __opts__[value]
    master_opts = __pillar__.get("master", {})
    if value in master_opts:
        return master_opts[value]
    if value in __pillar__:
        return __pillar__[value]


def _get_profile(service, region, key, keyid, profile):
    if profile:
        if isinstance(profile, str):
            _profile = _option(profile)
        elif isinstance(profile, dict):
            _profile = profile
        key = _profile.get("key", None)
        keyid = _profile.get("keyid", None)
        region = _profile.get("region", None)

    if not region and _option(service + ".region"):
        region = _option(service + ".region")

    if not region:
        region = "us-east-1"
        log.info("Assuming default region %s", region)

    if not key and _option(service + ".key"):
        key = _option(service + ".key")
    if not keyid and _option(service + ".keyid"):
        keyid = _option(service + ".keyid")

    label = "boto_{}:".format(service)
    if keyid:
        hash_string = region + keyid + key
        hash_string = salt.utils.stringutils.to_bytes(hash_string)
        cxkey = label + hashlib.md5(hash_string).hexdigest()
    else:
        cxkey = label + region

    return (cxkey, region, key, keyid)


def cache_id(
    service,
    name,
    sub_resource=None,
    resource_id=None,
    invalidate=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Cache, invalidate, or retrieve an AWS resource id keyed by name.

    .. code-block:: python

        __utils__['boto.cache_id']('ec2', 'myinstance',
                                   'i-a1b2c3',
                                   profile='custom_profile')
    """

    cxkey, _, _, _ = _get_profile(service, region, key, keyid, profile)
    if sub_resource:
        cxkey = "{}:{}:{}:id".format(cxkey, sub_resource, name)
    else:
        cxkey = "{}:{}:id".format(cxkey, name)

    if invalidate:
        if cxkey in __context__:
            del __context__[cxkey]
            return True
        elif resource_id in __context__.values():
            ctx = {k: v for k, v in __context__.items() if v != resource_id}
            __context__.clear()
            __context__.update(ctx)
            return True
        else:
            return False
    if resource_id:
        __context__[cxkey] = resource_id
        return True

    return __context__.get(cxkey)


def cache_id_func(service):
    """
    Returns a partial `cache_id` function for the provided service.

    .. code-block:: python

        cache_id = __utils__['boto.cache_id_func']('ec2')
        cache_id('myinstance', 'i-a1b2c3')
        instance_id = cache_id('myinstance')
    """
    return partial(cache_id, service)


def get_connection(
    service, module=None, region=None, key=None, keyid=None, profile=None
):
    """
    Return a boto connection for the service.

    .. code-block:: python

        conn = __utils__['boto.get_connection']('ec2', profile='custom_profile')
    """

    module = module or service

    cxkey, region, key, keyid = _get_profile(service, region, key, keyid, profile)
    cxkey = cxkey + ":conn3"

    if cxkey in __context__:
        return __context__[cxkey]

    try:
        session = boto3.session.Session(
            aws_access_key_id=keyid, aws_secret_access_key=key, region_name=region
        )
        if session is None:
            raise SaltInvocationError('Region "{}" is not valid.'.format(region))
        conn = session.client(module)
        if conn is None:
            raise SaltInvocationError('Region "{}" is not valid.'.format(region))
    except botocore.exceptions.NoCredentialsError:
        raise SaltInvocationError(
            "No authentication credentials found when "
            "attempting to make boto {} connection to "
            'region "{}".'.format(service, region)
        )
    __context__[cxkey] = conn
    return conn


def get_connection_func(service, module=None):
    """
    Returns a partial `get_connection` function for the provided service.

    .. code-block:: python

        get_conn = __utils__['boto.get_connection_func']('ec2')
        conn = get_conn()
    """
    return partial(get_connection, service, module=module)


def get_region(service, region, profile):
    """
    Retrieve the region for a particular AWS service based on configured region and/or profile.
    """
    _, region, _, _ = _get_profile(service, region, None, None, profile)

    return region


def get_error(e):
    # The returns from boto modules vary greatly between modules. We need to
    # assume that none of the data we're looking for exists.
    aws = {}

    message = ""
    message = e.args[0]

    r = {"message": message}
    if aws:
        r["aws"] = aws
    return r


def exactly_n(l, n=1):
    """
    Tests that exactly N items in an iterable are "truthy" (neither None,
    False, nor 0).
    """
    i = iter(l)
    return all(any(i) for j in range(n)) and not any(i)


def exactly_one(l):
    return exactly_n(l)


def assign_funcs(
    modname,
    service,
    module=None,
    get_conn_funcname="_get_conn",
    cache_id_funcname="_cache_id",
    exactly_one_funcname="_exactly_one",
):
    """
    Assign _get_conn and _cache_id functions to the named module.

    .. code-block:: python

        _utils__['boto.assign_partials'](__name__, 'ec2')
    """
    mod = sys.modules[modname]
    setattr(mod, get_conn_funcname, get_connection_func(service, module=module))
    setattr(mod, cache_id_funcname, cache_id_func(service))

    # TODO: Remove this and import salt.utils.data.exactly_one into boto_* modules instead
    # Leaving this way for now so boto modules can be back ported
    if exactly_one_funcname is not None:
        setattr(mod, exactly_one_funcname, exactly_one)


def paged_call(function, *args, **kwargs):
    """Retrieve full set of values from a boto3 API call that may truncate
    its results, yielding each page as it is obtained.
    """
    marker_flag = kwargs.pop("marker_flag", "NextMarker")
    marker_arg = kwargs.pop("marker_arg", "Marker")
    while True:
        ret = function(*args, **kwargs)
        marker = ret.get(marker_flag)
        yield ret
        if not marker:
            break
        kwargs[marker_arg] = marker


def ordered(obj):
    if isinstance(obj, (list, tuple)):
        return sorted(ordered(x) for x in obj)
    elif isinstance(obj, dict):
        return {str(k) if isinstance(k, str) else k: ordered(v) for k, v in obj.items()}
    elif isinstance(obj, str):
        return str(obj)
    return obj


def json_objs_equal(left, right):
    """Compare two parsed JSON objects, given non-ordering in JSON objects"""
    return ordered(left) == ordered(right)



import sys

import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_salt_state_file(salt_cli, salt_minion):
    """
    Test state file
    """
    if sys.platform.startswith("win"):
        ret = salt_cli.run("state.apply", "win_states", minion_tgt=salt_minion.id)
    else:
        ret = salt_cli.run("state.apply", "states", minion_tgt=salt_minion.id)

    assert ret.data, ret
    if ret.stdout and "Minion did not return" in ret.stdout:
        pytest.skip("Skipping test, state took too long to apply")
    sls_ret = ret.data[next(iter(ret.data))]
    assert "changes" in sls_ret
    assert "name" in sls_ret



"""
Management of MySQL grants (user permissions)
=============================================

:depends:   - MySQLdb Python module
:configuration: See :py:mod:`salt.modules.mysql` for setup instructions.

The mysql_grants module is used to grant and revoke MySQL permissions.

The ``name`` you pass in purely symbolic and does not have anything to do
with the grant itself.

The ``database`` parameter needs to specify a 'priv_level' in the same
specification as defined in the MySQL documentation:

* \\*
* \\*.\\*
* db_name.\\*
* db_name.tbl_name
* etc...

This state is not able to set password for the permission from the
specified host. See :py:mod:`salt.states.mysql_user` for further
instructions.

.. code-block:: yaml

   frank_exampledb:
      mysql_grants.present:
       - grant: select,insert,update
       - database: exampledb.*
       - user: frank
       - host: localhost

   frank_otherdb:
     mysql_grants.present:
       - grant: all privileges
       - database: otherdb.*
       - user: frank

   restricted_singletable:
     mysql_grants.present:
       - grant: select
       - database: somedb.sometable
       - user: joe
"""

import sys


def __virtual__():
    """
    Only load if the mysql module is available
    """
    if "mysql.grant_exists" in __salt__:
        return True
    return (False, "mysql module could not be loaded")


def _get_mysql_error():
    """
    Look in module context for a MySQL error. Eventually we should make a less
    ugly way of doing this.
    """
    return sys.modules[__salt__["test.ping"].__module__].__context__.pop(
        "mysql.error", None
    )


def present(
    name,
    grant=None,
    database=None,
    user=None,
    host="localhost",
    grant_option=False,
    escape=True,
    revoke_first=False,
    ssl_option=False,
    **connection_args
):
    """
    Ensure that the grant is present with the specified properties

    name
        The name (key) of the grant to add

    grant
        The grant priv_type (i.e. select,insert,update OR all privileges)

    database
        The database priv_level (i.e. db.tbl OR db.*)

    user
        The user to apply the grant to

    host
        The network/host that the grant should apply to

    grant_option
        Adds the WITH GRANT OPTION to the defined grant. Default is ``False``

    escape
        Defines if the database value gets escaped or not. Default is ``True``

    revoke_first
        By default, MySQL will not do anything if you issue a command to grant
        privileges that are more restrictive than what's already in place. This
        effectively means that you cannot downgrade permissions without first
        revoking permissions applied to a db.table/user pair first.

        To have Salt forcibly revoke perms before applying a new grant, enable
        the 'revoke_first options.

        WARNING: This will *remove* permissions for a database before attempting
        to apply new permissions. There is no guarantee that new permissions
        will be applied correctly which can leave your database security in an
        unknown and potentially dangerous state.
        Use with caution!

        Default is ``False``

    ssl_option
        Adds the specified ssl options for the connecting user as requirements for
        this grant. Value is a list of single-element dicts corresponding to the
        list of ssl options to use.

        Possible key/value pairings for the dicts in the value:

        .. code-block:: text

            - SSL: True
            - X509: True
            - SUBJECT: <subject>
            - ISSUER: <issuer>
            - CIPHER: <cipher>

        The non-boolean ssl options take a string as their values, which should
        be an appropriate value as specified by the MySQL documentation for these
        options.

        Default is ``False`` (no ssl options will be used)
    """
    comment = "Grant {0} on {1} to {2}@{3} is already present"
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": comment.format(grant, database, user, host),
    }
    # check if grant exists
    if __salt__["mysql.grant_exists"](
        grant, database, user, host, grant_option, escape, **connection_args
    ):
        return ret
    else:
        err = _get_mysql_error()
        if err is not None:
            ret["comment"] = err
            ret["result"] = False
            return ret
    if revoke_first and not __opts__["test"]:
        #  for each grant, break into tokens and see if its on the same
        # user/db/table as ours. (there is probably only one)
        user_grants = __salt__["mysql.user_grants"](user, host, **connection_args)
        if not user_grants:
            user_grants = []
        for user_grant in user_grants:
            token_grants = __salt__["mysql.tokenize_grant"](user_grant)
            db_part = database.rpartition(".")
            my_db = db_part[0]
            my_table = db_part[2]
            my_db = __salt__["mysql.quote_identifier"](my_db, (my_table == "*"))
            my_table = __salt__["mysql.quote_identifier"](my_table)
            # Removing per table grants in case of database level grant !!!
            if token_grants["database"] == my_db:
                grant_to_revoke = ",".join(token_grants["grant"]).rstrip(",")
                __salt__["mysql.grant_revoke"](
                    grant=grant_to_revoke,
                    database=database,
                    user=user,
                    host=host,
                    grant_option=grant_option,
                    escape=escape,
                    **connection_args
                )

    # The grant is not present, make it!
    if __opts__["test"]:
        # there is probably better things to make in test mode
        ret["result"] = None
        ret["comment"] = "MySQL grant {} is set to be created".format(name)
        return ret
    if __salt__["mysql.grant_add"](
        grant, database, user, host, grant_option, escape, ssl_option, **connection_args
    ):
        ret["comment"] = "Grant {0} on {1} to {2}@{3} has been added"
        ret["comment"] = ret["comment"].format(grant, database, user, host)
        ret["changes"][name] = "Present"
    else:
        ret["comment"] = 'Failed to execute: "GRANT {0} ON {1} TO {2}@{3}"'
        ret["comment"] = ret["comment"].format(grant, database, user, host)
        err = _get_mysql_error()
        if err is not None:
            ret["comment"] += " ({})".format(err)
        ret["result"] = False
    return ret


def absent(
    name,
    grant=None,
    database=None,
    user=None,
    host="localhost",
    grant_option=False,
    escape=True,
    **connection_args
):
    """
    Ensure that the grant is absent

    name
        The name (key) of the grant to add

    grant
        The grant priv_type (i.e. select,insert,update OR all privileges)

    database
        The database priv_level (i.e. db.tbl OR db.*)

    user
        The user to apply the grant to

    host
        The network/host that the grant should apply to
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # Check if grant exists, and if so, remove it
    if __salt__["mysql.grant_exists"](
        grant, database, user, host, grant_option, escape, **connection_args
    ):

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "MySQL grant {} is set to be revoked".format(name)
            return ret
        if __salt__["mysql.grant_revoke"](
            grant, database, user, host, grant_option, **connection_args
        ):
            ret["comment"] = "Grant {} on {} for {}@{} has been revoked".format(
                grant, database, user, host
            )
            ret["changes"][name] = "Absent"
            return ret
        else:
            err = _get_mysql_error()
            if err is not None:
                ret[
                    "comment"
                ] = "Unable to revoke grant {} on {} for {}@{} ({})".format(
                    grant, database, user, host, err
                )
                ret["result"] = False
                return ret
    else:
        err = _get_mysql_error()
        if err is not None:
            ret[
                "comment"
            ] = "Unable to determine if grant {} on {} for {}@{} exists ({})".format(
                grant, database, user, host, err
            )
            ret["result"] = False
            return ret

    # fallback
    ret[
        "comment"
    ] = "Grant {} on {} to {}@{} is not present, so it cannot be revoked".format(
        grant, database, user, host
    )
    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.win_groupadd as win_groupadd
import salt.utils.win_functions
from tests.support.helpers import TstSuiteLoggingHandler
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase

# pylint: disable=unused-import
try:
    import pythoncom
    import pywintypes
    import win32com

    PYWINTYPES_ERROR = pywintypes.com_error(
        -1234, "Exception occurred.", (0, None, "C", None, 0, -2147352567), None
    )
    HAS_WIN_LIBS = True
except ImportError:
    HAS_WIN_LIBS = False
# pylint: enable=unused-import


class MockMember:
    def __init__(self, name):
        self.ADSPath = name


class MockGroupObj:
    def __init__(self, ads_name, ads_users):
        self._members = [MockMember(x) for x in ads_users]
        self.Name = ads_name

    def members(self):
        return self._members

    def Add(self, name):
        """
        This should be a no-op unless we want to test raising an error, in
        which case this should be overridden in a subclass.
        """

    def Remove(self, name):
        """
        This should be a no-op unless we want to test raising an error, in
        which case this should be overridden in a subclass.
        """


@pytest.mark.skipif(
    not HAS_WIN_LIBS,
    reason="win_groupadd unit tests can only be run if win32com, pythoncom, and pywintypes are"
    " installed",
)
class WinGroupTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_groupadd
    """

    def setup_loader_modules(self):
        self.sam_mock = MagicMock(side_effect=lambda x: "HOST\\" + x)
        return {win_groupadd: {"__opts__": {"test": False}}}

    def test_add(self):
        """
        Test adding a new group
        """
        info = MagicMock(return_value=False)
        with patch.object(win_groupadd, "info", info), patch.object(
            win_groupadd, "_get_computer_object", Mock()
        ):
            self.assertTrue(win_groupadd.add("foo"))

    def test_add_group_exists(self):
        """
        Test adding a new group if the group already exists
        """
        info = MagicMock(
            return_value={
                "name": "foo",
                "passwd": None,
                "gid": None,
                "members": ["HOST\\spongebob"],
            }
        )
        with patch.object(win_groupadd, "info", info), patch.object(
            win_groupadd, "_get_computer_object", Mock()
        ):
            self.assertFalse(win_groupadd.add("foo"))

    def test_add_error(self):
        """
        Test adding a group and encountering an error
        """

        class CompObj:
            def Create(self, type, name):
                raise PYWINTYPES_ERROR

        obj_comp_mock = MagicMock(return_value=CompObj())
        info = MagicMock(return_value=False)
        with patch.object(win_groupadd, "info", info), patch.object(
            win_groupadd, "_get_computer_object", obj_comp_mock
        ):
            self.assertFalse(win_groupadd.add("foo"))

    def test_delete(self):
        """
        Test removing a group
        """
        info = MagicMock(
            return_value={
                "name": "foo",
                "passwd": None,
                "gid": None,
                "members": ["HOST\\spongebob"],
            }
        )
        with patch.object(win_groupadd, "info", info), patch.object(
            win_groupadd, "_get_computer_object", Mock()
        ):
            self.assertTrue(win_groupadd.delete("foo"))

    def test_delete_no_group(self):
        """
        Test removing a group that doesn't exists
        """
        info = MagicMock(return_value=False)
        with patch.object(win_groupadd, "info", info), patch.object(
            win_groupadd, "_get_computer_object", Mock()
        ):
            self.assertFalse(win_groupadd.delete("foo"))

    def test_delete_error(self):
        """
        Test removing a group and encountering an error
        """

        class CompObj:
            def Delete(self, type, name):
                raise PYWINTYPES_ERROR

        obj_comp_mock = MagicMock(return_value=CompObj())
        info = MagicMock(
            return_value={
                "name": "foo",
                "passwd": None,
                "gid": None,
                "members": ["HOST\\spongebob"],
            }
        )
        with patch.object(win_groupadd, "info", info), patch.object(
            win_groupadd, "_get_computer_object", obj_comp_mock
        ):
            self.assertFalse(win_groupadd.delete("foo"))

    def test_info(self):
        """
        Test if it return information about a group.
        """
        obj_group_mock = MagicMock(
            return_value=MockGroupObj("salt", ["WinNT://HOST/steve"])
        )
        with patch.object(win_groupadd, "_get_group_object", obj_group_mock):
            self.assertDictEqual(
                win_groupadd.info("salt"),
                {
                    "gid": None,
                    "members": ["HOST\\steve"],
                    "passwd": None,
                    "name": "salt",
                },
            )

    def test_getent(self):
        obj_group_mock = MagicMock(
            return_value=[
                MockGroupObj("salt", ["WinNT://HOST/steve"]),
                MockGroupObj("salty", ["WinNT://HOST/spongebob"]),
            ]
        )
        mock_g_to_g = MagicMock(side_effect=[1, 2])
        with patch.object(win_groupadd, "_get_all_groups", obj_group_mock), patch.dict(
            win_groupadd.__salt__, {"file.group_to_gid": mock_g_to_g}
        ):
            self.assertListEqual(
                win_groupadd.getent(),
                [
                    {
                        "gid": 1,
                        "members": ["HOST\\steve"],
                        "name": "salt",
                        "passwd": "x",
                    },
                    {
                        "gid": 2,
                        "members": ["HOST\\spongebob"],
                        "name": "salty",
                        "passwd": "x",
                    },
                ],
            )

    def test_getent_context(self):
        """
        Test group.getent is using the values in __context__
        """
        with patch.dict(win_groupadd.__context__, {"group.getent": True}):
            self.assertTrue(win_groupadd.getent())

    def test_adduser(self):
        """
        Test adding a user to a group
        """
        obj_group_mock = MagicMock(
            return_value=MockGroupObj("foo", ["WinNT://HOST/steve"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertTrue(win_groupadd.adduser("foo", "spongebob"))

    def test_adduser_already_exists(self):
        """
        Test adding a user that already exists
        """
        obj_group_mock = MagicMock(
            return_value=MockGroupObj("foo", ["WinNT://HOST/steve"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertFalse(win_groupadd.adduser("foo", "steve"))

    def test_adduser_error(self):
        """
        Test adding a user and encountering an error
        """
        msg = "An unknown directory object was requested"
        error = pywintypes.com_error(
            -1234, "Exception occurred.", (0, None, msg, None, 0, -2147352567), None
        )

        # Create mock group object with mocked Add function which raises the
        # exception we need in order to test the error case.
        class GroupObj(MockGroupObj):
            def Add(self, name):
                raise error

        obj_group_mock = MagicMock(return_value=GroupObj("foo", ["WinNT://HOST/steve"]))
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            with TstSuiteLoggingHandler() as handler:
                self.assertFalse(win_groupadd.adduser("foo", "username"))
                expected = (
                    "ERROR:Failed to add HOST\\username to group foo. An unknown"
                    " directory object was requested"
                )
                self.assertIn(expected, handler.messages)

    def test_adduser_group_does_not_exist(self):
        obj_group_mock = MagicMock(side_effect=PYWINTYPES_ERROR)
        with patch.object(win_groupadd, "_get_group_object", obj_group_mock):
            self.assertFalse(win_groupadd.adduser("foo", "spongebob"))

    def test_deluser(self):
        """
        Test removing a user from a group
        """
        # Test removing a user
        obj_group_mock = MagicMock(
            return_value=MockGroupObj("foo", ["WinNT://HOST/spongebob"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertTrue(win_groupadd.deluser("foo", "spongebob"))

    def test_deluser_no_user(self):
        """
        Test removing a user from a group and that user is not a member of the
        group
        """
        obj_group_mock = MagicMock(
            return_value=MockGroupObj("foo", ["WinNT://HOST/steve"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertFalse(win_groupadd.deluser("foo", "spongebob"))

    def test_deluser_error(self):
        """
        Test removing a user and encountering an error
        """

        class GroupObj(MockGroupObj):
            def Remove(self, name):
                raise PYWINTYPES_ERROR

        obj_group_mock = MagicMock(
            return_value=GroupObj("foo", ["WinNT://HOST/spongebob"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertFalse(win_groupadd.deluser("foo", "spongebob"))

    def test_deluser_group_does_not_exist(self):
        obj_group_mock = MagicMock(side_effect=PYWINTYPES_ERROR)
        with patch.object(win_groupadd, "_get_group_object", obj_group_mock):
            self.assertFalse(win_groupadd.deluser("foo", "spongebob"))

    def test_members(self):
        """
        Test adding a list of members to a group, all existing users removed
        """
        obj_group_mock = MagicMock(
            return_value=MockGroupObj("foo", ["WinNT://HOST/steve"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertTrue(win_groupadd.members("foo", "spongebob,patrick,squidward"))
            obj_group_mock.assert_called_once_with("foo")

    def test_members_correct_membership(self):
        """
        Test adding a list of users where the list of users already exists
        """
        members_list = [
            "WinNT://HOST/spongebob",
            "WinNT://HOST/squidward",
            "WinNT://HOST/patrick",
        ]
        obj_group_mock = MagicMock(return_value=MockGroupObj("foo", members_list))
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertTrue(win_groupadd.members("foo", "spongebob,patrick,squidward"))
            obj_group_mock.assert_called_once_with("foo")

    def test_members_group_does_not_exist(self):
        """
        Test adding a list of users where the group does not exist
        """
        obj_group_mock = MagicMock(side_effect=PYWINTYPES_ERROR)
        with patch.object(
            salt.utils.win_functions, "get_sam_name", self.sam_mock
        ), patch.object(win_groupadd, "_get_group_object", obj_group_mock):
            self.assertFalse(win_groupadd.members("foo", "spongebob"))

    def test_members_fail_to_remove(self):
        """
        Test adding a list of members and fail to remove members not in the list
        """

        class GroupObj(MockGroupObj):
            def Remove(self, name):
                raise PYWINTYPES_ERROR

        obj_group_mock = MagicMock(
            return_value=GroupObj("foo", ["WinNT://HOST/spongebob"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertFalse(win_groupadd.members("foo", "patrick"))
            obj_group_mock.assert_called_once_with("foo")

    def test_members_fail_to_add(self):
        """
        Test adding a list of members and failing to add
        """

        class GroupObj(MockGroupObj):
            def Add(self, name):
                raise PYWINTYPES_ERROR

        obj_group_mock = MagicMock(
            return_value=GroupObj("foo", ["WinNT://HOST/spongebob"])
        )
        with patch.object(
            win_groupadd, "_get_group_object", obj_group_mock
        ), patch.object(salt.utils.win_functions, "get_sam_name", self.sam_mock):
            self.assertFalse(win_groupadd.members("foo", "patrick"))
            obj_group_mock.assert_called_once_with("foo")

    def test_list_groups(self):
        """
        Test that list groups returns a list of groups by name
        """
        obj_group_mock = MagicMock(
            return_value=[
                MockGroupObj("salt", ["WinNT://HOST/steve"]),
                MockGroupObj("salty", ["WinNT://HOST/Administrator"]),
            ]
        )
        with patch.object(win_groupadd, "_get_all_groups", obj_group_mock):
            self.assertListEqual(win_groupadd.list_groups(), ["salt", "salty"])

    def test_list_groups_context(self):
        """
        Test group.list_groups is using the values in __context__
        """
        with patch.dict(win_groupadd.__context__, {"group.list_groups": True}):
            self.assertTrue(win_groupadd.list_groups())

    def test__get_computer_object(self):
        """
        Test the _get_computer_object function. Should return a Com object, but
        we're going to mock just an object
        """
        with patch("salt.utils.winapi.Com", MagicMock()), patch(
            "win32com.client.Dispatch", autospec=True
        ) as com_obj:
            com_obj.GetObject = [object]
            test = win_groupadd._get_computer_object()
            assert isinstance(test, object)

    def test__get_group_object(self):
        """
        Test the _get_group_object function. Should return a Com object, but
        we're going to mock just an object
        """
        with patch("salt.utils.winapi.Com", MagicMock()), patch(
            "win32com.client.Dispatch", autospec=True
        ) as com_obj:
            com_obj.GetObject = [object]
            test = win_groupadd._get_group_object("steve")
            assert isinstance(test, object)

    def test__get_all_groups(self):
        """
        Test the _get_all_groups function. It doesn't actually return a list.
        It returns a Com Object that is an iterator, so we're just going to mock
        an object
        """
        with patch("salt.utils.winapi.Com", MagicMock()), patch(
            "win32com.client.Dispatch", autospec=True
        ) as com_obj:
            com_obj.GetObject = [object]
            test = win_groupadd._get_all_groups()
            assert isinstance(test, object)



import logging

from salt._logging.handlers import (
    FileHandler,
    QueueHandler,
    RotatingFileHandler,
    StreamHandler,
    SysLogHandler,
    TemporaryLoggingHandler,
    WatchedFileHandler,
)
from salt.utils.versions import warn_until_date

warn_until_date(
    "20240101",
    "Please stop using '{name}' and instead use 'salt._logging.handlers'. "
    "'{name}' will go away after {{date}}.".format(name=__name__),
)

NullHandler = logging.NullHandler



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.boto_iam_role as boto_iam_role
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_iam_role: {}}


def test_present():
    """
    Test to ensure the IAM role exists.
    """
    name = "myrole"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    _desc_role = {
        "create_date": "2015-02-11T19:47:14Z",
        "role_id": "HIUHBIUBIBNKJNBKJ",
        "assume_role_policy_document": {
            "Version": "2008-10-17",
            "Statement": [
                {
                    "Action": "sts:AssumeRole",
                    "Principal": {"Service": "ec2.amazonaws.com"},
                    "Effect": "Allow",
                }
            ],
        },
        "role_name": "myfakerole",
        "path": "/",
        "arn": "arn:aws:iam::12345:role/myfakerole",
    }
    _desc_role2 = {
        "create_date": "2015-02-11T19:47:14Z",
        "role_id": "HIUHBIUBIBNKJNBKJ",
        "assume_role_policy_document": {
            "Version": "2008-10-17",
            "Statement": [
                {
                    "Action": "sts:AssumeRole",
                    "Principal": {
                        "Service": ["ec2.amazonaws.com", "datapipeline.amazonaws.com"]
                    },
                    "Effect": "Allow",
                }
            ],
        },
        "role_name": "myfakerole",
        "path": "/",
        "arn": "arn:aws:iam::12345:role/myfakerole",
    }
    mock_desc = MagicMock(
        side_effect=[False, _desc_role, _desc_role, _desc_role2, _desc_role]
    )
    _build_policy = {
        "Version": "2008-10-17",
        "Statement": [
            {
                "Action": "sts:AssumeRole",
                "Effect": "Allow",
                "Principal": {"Service": "ec2.amazonaws.com"},
            }
        ],
    }
    mock_policy = MagicMock(return_value=_build_policy)
    mock_ipe = MagicMock(side_effect=[False, True, True, True])
    mock_pa = MagicMock(side_effect=[False, True, True, True])
    mock_bool = MagicMock(return_value=False)
    mock_lst = MagicMock(return_value=[])
    with patch.dict(
        boto_iam_role.__salt__,
        {
            "boto_iam.describe_role": mock_desc,
            "boto_iam.create_role": mock_bool,
            "boto_iam.build_policy": mock_policy,
            "boto_iam.update_assume_role_policy": mock_bool,
            "boto_iam.instance_profile_exists": mock_ipe,
            "boto_iam.list_attached_role_policies": mock_lst,
            "boto_iam.create_instance_profile": mock_bool,
            "boto_iam.profile_associated": mock_pa,
            "boto_iam.associate_profile_to_role": mock_bool,
            "boto_iam.list_role_policies": mock_lst,
        },
    ):
        with patch.dict(boto_iam_role.__opts__, {"test": False}):
            comt = " Failed to create {} IAM role.".format(name)
            ret.update({"comment": comt})
            assert boto_iam_role.present(name) == ret

            comt = " myrole role present. Failed to create myrole instance profile."
            ret.update({"comment": comt})
            assert boto_iam_role.present(name) == ret

            comt = (
                " myrole role present.  Failed to associate myrole"
                " instance profile with myrole role."
            )
            ret.update({"comment": comt})
            assert boto_iam_role.present(name) == ret

            comt = " myrole role present. Failed to update assume role policy."
            ret.update({"comment": comt})

            assert boto_iam_role.present(name) == ret
            comt = " myrole role present.    "
            ret.update({"comment": comt, "result": True})
            assert boto_iam_role.present(name) == ret


def test_absent():
    """
    Test to ensure the IAM role is deleted.
    """
    name = "myrole"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(
        side_effect=[
            ["mypolicy"],
            ["mypolicy"],
            False,
            True,
            False,
            False,
            True,
            False,
            False,
            False,
            True,
        ]
    )
    mock_bool = MagicMock(return_value=False)
    mock_lst = MagicMock(return_value=[])
    with patch.dict(
        boto_iam_role.__salt__,
        {
            "boto_iam.list_role_policies": mock,
            "boto_iam.delete_role_policy": mock_bool,
            "boto_iam.profile_associated": mock,
            "boto_iam.disassociate_profile_from_role": mock_bool,
            "boto_iam.instance_profile_exists": mock,
            "boto_iam.list_attached_role_policies": mock_lst,
            "boto_iam.delete_instance_profile": mock_bool,
            "boto_iam.role_exists": mock,
            "boto_iam.delete_role": mock_bool,
        },
    ):
        with patch.dict(boto_iam_role.__opts__, {"test": False}):
            comt = " Failed to add policy mypolicy to role myrole"
            ret.update(
                {
                    "comment": comt,
                    "changes": {
                        "new": {"policies": ["mypolicy"]},
                        "old": {"policies": ["mypolicy"]},
                    },
                }
            )
            assert boto_iam_role.absent(name) == ret

            comt = (
                " No policies in role myrole."
                " No attached policies in role myrole. Failed to disassociate "
                "myrole instance profile from myrole role."
            )
            ret.update({"comment": comt, "changes": {}})
            assert boto_iam_role.absent(name) == ret

            comt = (
                " No policies in role myrole."
                " No attached policies in role myrole. "
                " Failed to delete myrole instance profile."
            )
            ret.update({"comment": comt, "changes": {}})
            assert boto_iam_role.absent(name) == ret

            comt = (
                " No policies in role myrole."
                " No attached policies in role myrole.  myrole instance profile "
                "does not exist. Failed to delete myrole iam role."
            )
            ret.update({"comment": comt, "changes": {}})
            assert boto_iam_role.absent(name) == ret



import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


def test_ssh_raw(salt_ssh_cli):
    """
    test salt-ssh with -r argument
    """
    msg = "password: foo"
    ret = salt_ssh_cli.run("--raw", "echo", msg, _timeout=60)
    assert ret.returncode == 0
    assert ret.data
    assert "retcode" in ret.data
    assert ret.data["retcode"] == 0
    assert "stdout" in ret.data
    assert ret.data["stdout"] == msg + "\n"



"""
Unit tests for chocolatey state
"""
import logging

import pytest

import salt.modules.chocolatey as chocolatey_mod
import salt.states.chocolatey as chocolatey
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def choco_path():
    return "C:\\path\\to\\chocolatey.exe"


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        chocolatey: {
            "__opts__": minion_opts,
            "__salt__": {},
            "__context__": {},
        },
        chocolatey_mod: {
            "__opts__": minion_opts,
            "__context__": {},
        },
    }


@pytest.fixture(scope="module")
def pkgs():
    return {
        "pkga": {"old": "1.0.1", "new": "2.0.1"},
        "pkgb": {"old": "1.0.2", "new": "2.0.2"},
        "pkgc": {"old": "1.0.3", "new": "2.0.3"},
    }


@pytest.fixture(scope="module")
def list_sources():
    _ret = {
        "chocolatey": {
            "URL": "https://community.chocolatey.org/api/v2/",
            "Disabled": False,
            "User": "user",
        },
        "community": {
            "URL": "https://community.chocolatey.org/api/v2/",
            "Disabled": False,
            "User": "user",
        },
    }
    return _ret


def test_source_present(list_sources):
    """
    Test chocolatey.source_present with simulated changes
    """

    before_list_sources = {
        "chocolatey": {
            "URL": "https://community.chocolatey.org/api/v2/",
            "Disabled": False,
            "User": "user",
        }
    }

    list_sources_sideeffect = MagicMock(side_effect=[before_list_sources, list_sources])

    with patch.dict(
        chocolatey.__salt__,
        {
            "chocolatey.list_sources": list_sources_sideeffect,
            "chocolatey.add_source": chocolatey_mod.add_source,
        },
    ):

        # Run state with test=true
        stdout_ret = (
            "Added community - https://community.chocolatey.org/api/v2/ (Priority 5)"
        )
        cmd_run_all_mock = MagicMock(return_value={"retcode": 0, "stdout": stdout_ret})
        cmd_run_which_mock = MagicMock(return_value=choco_path)
        with patch.dict(
            chocolatey_mod.__salt__,
            {
                "cmd.which": cmd_run_which_mock,
                "cmd.run_all": cmd_run_all_mock,
            },
        ):
            ret = chocolatey.source_present(
                "community",
                source_location="https://community.chocolatey.org/api/v2/",
                username="username",
                password="password",
                priority="5",
            )
            assert ret["result"] is True
            assert ret["name"] == "community"
            assert ret["comment"] == "Source community added successfully"
            assert ret["changes"] == {
                "community": {
                    "old": "",
                    "new": {
                        "URL": "https://community.chocolatey.org/api/v2/",
                        "Disabled": False,
                        "User": "user",
                    },
                }
            }



"""
Tests for the service state
"""

import os

import pytest

import salt.utils.path
import salt.utils.platform

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


INIT_DELAY = 5
STOPPED = False
RUNNING = True


@pytest.fixture
def service_name(grains, modules):
    # For local testing purposes
    env_name = os.environ.get("SALT_FUNCTIONAL_TEST_SERVICE_NAME")
    if env_name is not None:
        return env_name

    service_name = "cron"
    cmd_name = "crontab"
    os_family = grains.get("os_family")
    is_systemd = grains.get("systemd")
    if os_family == "RedHat":
        service_name = "crond"
    elif os_family == "Arch":
        service_name = "sshd"
        cmd_name = "systemctl"
    elif os_family == "MacOS":
        service_name = "com.apple.AirPlayXPCHelper"
    elif os_family == "Windows":
        service_name = "Spooler"

    if os_family != "Windows" and salt.utils.path.which(cmd_name) is None:
        pytest.skip("{} is not installed".format(cmd_name))

    if is_systemd and modules.service.offline():
        pytest.skip("systemd is OFFLINE")

    return service_name


@pytest.fixture(autouse=True)
def setup_service(service_name, modules):
    pre_srv_status = modules.service.status(service_name)
    pre_srv_enabled = service_name in modules.service.get_enabled()

    try:
        yield pre_srv_status
    finally:
        post_srv_status = modules.service.status(service_name)
        post_srv_enabled = service_name in modules.service.get_enabled()

        if post_srv_status != pre_srv_status:
            if pre_srv_status:
                modules.service.start(service_name)
            else:
                modules.service.stop(service_name)

        if post_srv_enabled != pre_srv_enabled:
            if pre_srv_enabled:
                modules.service.enable(service_name)
            else:
                modules.service.disable(service_name)


def check_service_status(exp_return, modules, service_name):
    """
    helper method to check status of service
    """
    check_status = modules.service.status(service_name)

    if check_status is not exp_return:
        pytest.fail("status of service is not returning correctly")


@pytest.mark.slow_test
def test_service_running(service_name, modules, states):
    """
    test service.running state module
    """
    if modules.service.status(service_name):
        stop_service = modules.service.stop(service_name)
        assert stop_service is True
    check_service_status(STOPPED, modules, service_name)

    if salt.utils.platform.is_darwin():
        # make sure the service is enabled on macosx
        enable = modules.service.enable(service_name)

    start_service = states.service.running(service_name)
    assert start_service.full_return["result"] is True
    check_service_status(RUNNING, modules, service_name)


@pytest.mark.slow_test
def test_service_dead(service_name, modules, states):
    """
    test service.dead state module
    """
    start_service = states.service.running(service_name)
    assert start_service.full_return["result"] is True
    check_service_status(RUNNING, modules, service_name)

    ret = states.service.dead(service_name)
    assert ret.full_return["result"] is True
    check_service_status(STOPPED, modules, service_name)


@pytest.mark.slow_test
def test_service_dead_init_delay(service_name, modules, states):
    """
    test service.dead state module
    """
    start_service = states.service.running(service_name)
    assert start_service.full_return["result"] is True
    check_service_status(RUNNING, modules, service_name)

    ret = states.service.dead(service_name, init_delay=INIT_DELAY)
    assert ret.full_return["result"] is True
    check_service_status(STOPPED, modules, service_name)



"""
Return salt data via Slack using Incoming Webhooks

:codeauthor: `Carlos D. Álvaro <github@cdalvaro.io>`

The following fields can be set in the minion conf file:

.. code-block:: yaml

    slack_webhook.webhook (required, the webhook id. Just the part after: 'https://hooks.slack.com/services/')
    slack_webhook.success_title (optional, short title for succeeded states. By default: '{id} | Succeeded')
    slack_webhook.failure_title (optional, short title for failed states. By default: '{id} | Failed')
    slack_webhook.author_icon (optional, a URL that with a small 16x16px image. Must be of type: GIF, JPEG, PNG, and BMP)
    slack_webhook.show_tasks (optional, show identifiers for changed and failed tasks. By default: False)

Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location:

.. code-block:: yaml

    slack_webhook.webhook
    slack_webhook.success_title
    slack_webhook.failure_title
    slack_webhook.author_icon
    slack_webhook.show_tasks

Slack settings may also be configured as:

.. code-block:: yaml

    slack_webhook:
      webhook: T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX
      success_title: '[{id}] | Success'
      failure_title: '[{id}] | Failure'
      author_icon: https://platform.slack-edge.com/img/default_application_icon.png
      show_tasks: true

    alternative.slack_webhook:
      webhook: T00000000/C00000000/YYYYYYYYYYYYYYYYYYYYYYYY
      show_tasks: false

To use the Slack returner,
append '--return slack_webhook' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return slack_webhook

To use the alternative configuration,
append '--return_config alternative' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return slack_webhook --return_config alternative

"""

import json
import logging
import urllib.parse

import salt.returners
import salt.utils.http
import salt.utils.yaml

log = logging.getLogger(__name__)

__virtualname__ = "slack_webhook"

UNCHANGED_KEY = "unchanged"
CHANGED_KEY = "changed"
FAILED_KEY = "failed"
TASKS_KEY = "tasks"
COUNTER_KEY = "counter"
DURATION_KEY = "duration"
TOTAL_KEY = "total"


def _get_options(ret=None):
    """
    Get the slack_webhook options from salt.
    :param ret: Salt return dictionary
    :return: A dictionary with options
    """

    defaults = {
        "success_title": "{id} | Succeeded",
        "failure_title": "{id} | Failed",
        "author_icon": "",
        "show_tasks": False,
    }

    attrs = {
        "webhook": "webhook",
        "success_title": "success_title",
        "failure_title": "failure_title",
        "author_icon": "author_icon",
        "show_tasks": "show_tasks",
    }

    _options = salt.returners.get_returner_options(
        __virtualname__,
        ret,
        attrs,
        __salt__=__salt__,
        __opts__=__opts__,
        defaults=defaults,
    )
    return _options


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """

    return __virtualname__


def _sprinkle(config_str):
    """
    Sprinkle with grains of salt, that is
    convert "test {id} test {host} " types of strings
    :param config_str: The string to be sprinkled
    :return: The string sprinkled
    """

    parts = [x for sub in config_str.split("{") for x in sub.split("}")]
    for i in range(1, len(parts), 2):
        parts[i] = str(__grains__.get(parts[i], ""))
    return "".join(parts)


def _format_task(task):
    """
    Return a dictionary with the task ready for slack fileds
    :param task: The name of the task

    :return: A dictionary ready to be inserted in Slack fields array
    """

    return {"value": task, "short": False}


def _generate_payload(author_icon, title, report, **kwargs):
    """
    Prepare the payload for Slack
    :param author_icon: The url for the thumbnail to be displayed
    :param title: The title of the message
    :param report: A dictionary with the report of the Salt function
    :return: The payload ready for Slack
    """

    event_rtn = kwargs.get("event_rtn", False)

    if event_rtn is True:
        author_name = report["id"]
    else:
        author_name = _sprinkle("{id}")

    title = _sprinkle(title)

    text = "Function: {}\n".format(report.get("function"))
    if len(report.get("arguments", [])) > 0:
        text += "Function Args: {}\n".format(str(list(map(str, report["arguments"]))))

    text += "JID: {}\n".format(report.get("jid"))

    if TOTAL_KEY in report:
        text += "Total: {}\n".format(report[TOTAL_KEY])

    if DURATION_KEY in report:
        text += "Duration: {:.2f} secs".format(float(report[DURATION_KEY]))

    attachments = [
        {
            "fallback": title,
            "color": "#272727",
            "author_name": author_name,
            "author_link": _sprinkle("{localhost}"),
            "author_icon": author_icon,
            "title": "Success: {}".format(str(report["success"])),
            "text": text,
        }
    ]

    if UNCHANGED_KEY in report:
        # Unchanged
        attachments.append(
            {
                "color": "good",
                "title": "Unchanged: {}".format(
                    report[UNCHANGED_KEY].get(COUNTER_KEY, 0)
                ),
            }
        )

        # Changed
        changed = {
            "color": "warning",
            "title": "Changed: {}".format(report[CHANGED_KEY].get(COUNTER_KEY, 0)),
        }

        if len(report[CHANGED_KEY].get(TASKS_KEY, [])) > 0:
            changed["fields"] = list(map(_format_task, report[CHANGED_KEY][TASKS_KEY]))

        attachments.append(changed)

        # Failed
        failed = {
            "color": "danger",
            "title": "Failed: {}".format(report[FAILED_KEY].get(COUNTER_KEY, None)),
        }

        if len(report[FAILED_KEY].get(TASKS_KEY, [])) > 0:
            failed["fields"] = list(map(_format_task, report[FAILED_KEY][TASKS_KEY]))

        attachments.append(failed)

    else:
        attachments.append(
            {
                "color": "good" if report["success"] else "danger",
                "title": "Return: {}".format(report.get("return", None)),
            }
        )

    payload = {"attachments": attachments}

    return payload


def _process_state(returns):
    """
    Process the received output state
    :param returns A dictionary with the returns of the recipe
    :return A dictionary with Unchanges, Changed and Failed tasks
    """

    sorted_data = sorted(returns.items(), key=lambda s: s[1].get("__run_num__", 0))

    n_total = 0
    n_failed = 0
    n_changed = 0
    duration = 0.0

    changed_tasks = []
    failed_tasks = []

    # gather stats
    for state, data in sorted_data:
        # state: module, stateid, name, function
        _, stateid, _, _ = state.split("_|-")
        task = "{filename}.sls | {taskname}".format(
            filename=str(data.get("__sls__")), taskname=stateid
        )

        if not data.get("result", True):
            n_failed += 1
            failed_tasks.append(task)

        if data.get("changes", {}):
            n_changed += 1
            changed_tasks.append(task)

        n_total += 1
        try:
            duration += float(data.get("duration", 0.0))
        except ValueError:
            pass

    n_unchanged = n_total - n_failed - n_changed

    return {
        TOTAL_KEY: n_total,
        UNCHANGED_KEY: {COUNTER_KEY: n_unchanged},
        CHANGED_KEY: {COUNTER_KEY: n_changed, TASKS_KEY: changed_tasks},
        FAILED_KEY: {COUNTER_KEY: n_failed, TASKS_KEY: failed_tasks},
        DURATION_KEY: duration / 1000,
    }


def _state_return(ret):
    """
    Return True if ret is a Salt state return
    :param ret: The Salt return
    """

    ret_data = ret.get("return")
    if not isinstance(ret_data, dict):
        return False

    return ret_data and "__id__" in next(iter(ret_data.values()))


def _generate_report(ret, show_tasks):
    """
    Generate a report of the Salt function
    :param ret: The Salt return
    :param show_tasks: Flag to show the name of the changed and failed states
    :return: The report
    """

    report = {
        "id": ret.get("id"),
        "success": True if ret.get("retcode", 1) == 0 else False,
        "function": ret.get("fun"),
        "arguments": ret.get("fun_args", []),
        "jid": ret.get("jid"),
    }

    ret_return = ret.get("return")
    if _state_return(ret):
        ret_return = _process_state(ret_return)
        if not show_tasks:
            del ret_return[CHANGED_KEY][TASKS_KEY]
            del ret_return[FAILED_KEY][TASKS_KEY]
    elif isinstance(ret_return, dict):
        ret_return = {
            "return": "\n{}".format(salt.utils.yaml.safe_dump(ret_return, indent=2))
        }
    else:
        ret_return = {"return": ret_return}

    report.update(ret_return)

    return report


def _post_message(webhook, author_icon, title, report, **kwargs):
    """
    Send a message to a Slack room through a webhook
    :param webhook:     The url of the incoming webhook
    :param author_icon: The thumbnail image to be displayed on the right side of the message
    :param title:       The title of the message
    :param report:      The report of the function state
    :return:            Boolean if message was sent successfully
    """

    event_rtn = kwargs.get("event_rtn", False)

    payload = _generate_payload(author_icon, title, report, event_rtn=event_rtn)

    data = urllib.parse.urlencode({"payload": json.dumps(payload, ensure_ascii=False)})

    webhook_url = urllib.parse.urljoin("https://hooks.slack.com/services/", webhook)
    query_result = salt.utils.http.query(webhook_url, "POST", data=data)

    # Sometimes the status is not available, so status 200 is assumed when it is not present
    if (
        query_result.get("body", "failed") == "ok"
        and query_result.get("status", 200) == 200
    ):
        return True
    else:
        log.error("Slack incoming webhook message post result: %s", query_result)
        return {"res": False, "message": query_result.get("body", query_result)}


def returner(ret, **kwargs):
    """
    Send a slack message with the data through a webhook
    :param ret: The Salt return
    :return: The result of the post
    """

    event_rtn = kwargs.get("event_rtn", False)

    _options = _get_options(ret)

    webhook = _options.get("webhook", None)
    show_tasks = _options.get("show_tasks")
    author_icon = _options.get("author_icon")

    if not webhook or webhook == "":
        log.error("%s.webhook not defined in salt config", __virtualname__)
        return

    report = _generate_report(ret, show_tasks)

    if report.get("success"):
        title = _options.get("success_title")
    else:
        title = _options.get("failure_title")

    slack = _post_message(webhook, author_icon, title, report, event_rtn=event_rtn)

    return slack


def event_return(events):
    """
    Send event data to returner function
    :param events: The Salt event return
    :return: The result of the post
    """

    results = None

    for event in events:
        ret = event.get("data", False)

        if (
            ret
            and "saltutil.find_job" not in ret["fun"]
            or "salt/auth" not in ret["tag"]
        ):
            results = returner(ret, event_rtn=True)

    return results



"""
Enforce state for SSL/TLS
=========================

"""


import datetime
import logging
import time

__virtualname__ = "tls"
log = logging.getLogger(__name__)


def __virtual__():
    if "tls.cert_info" not in __salt__:
        return (False, "tls module could not be loaded")

    return __virtualname__


def valid_certificate(name, weeks=0, days=0, hours=0, minutes=0, seconds=0):
    """
    Verify that a TLS certificate is valid now and (optionally) will be valid
    for the time specified through weeks, days, hours, minutes, and seconds.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    now = time.time()
    try:
        cert_info = __salt__["tls.cert_info"](name)
    except OSError as exc:
        ret["comment"] = "{}".format(exc)
        ret["result"] = False
        log.error(ret["comment"])
        return ret

    # verify that the cert is valid *now*
    if now < cert_info["not_before"]:
        ret["comment"] = "Certificate is not yet valid"
        return ret
    if now > cert_info["not_after"]:
        ret["comment"] = "Certificate is expired"
        return ret

    # verify the cert will be valid for defined time
    delta_remaining = datetime.timedelta(seconds=cert_info["not_after"] - now)
    delta_kind_map = {
        "weeks": weeks,
        "days": days,
        "hours": hours,
        "minutes": minutes,
        "seconds": seconds,
    }

    delta_min = datetime.timedelta(**delta_kind_map)
    # if ther eisn't enough time remaining, we consider it a failure
    if delta_remaining < delta_min:
        ret["comment"] = "Certificate will expire in {}, which is less than {}".format(
            delta_remaining, delta_min
        )
        return ret

    ret["result"] = True
    ret["comment"] = "Certificate is valid for {}".format(delta_remaining)
    return ret



"""
Beacon to fire events at specific log messages.

.. versionadded:: 2017.7.0

"""
import logging

import salt.utils.beacons
import salt.utils.files
import salt.utils.platform

try:
    import re

    HAS_REGEX = True
except ImportError:
    HAS_REGEX = False

__virtualname__ = "log"
LOC_KEY = "log.loc"

SKEL = {}
SKEL["tag"] = ""
SKEL["match"] = "no"
SKEL["raw"] = ""
SKEL["error"] = ""


log = logging.getLogger(__name__)


def __virtual__():
    if not salt.utils.platform.is_windows() and HAS_REGEX:
        return __virtualname__
    err_msg = "Not available for Windows systems or when regex library is missing."
    log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
    return False, err_msg


def _get_loc():
    """
    return the active file location
    """
    if LOC_KEY in __context__:
        return __context__[LOC_KEY]


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for log beacon should be a list of dicts
    if not isinstance(config, list):
        return False, "Configuration for log beacon must be a list."

    config = salt.utils.beacons.list_to_dict(config)

    if "file" not in config:
        return False, "Configuration for log beacon must contain file option."
    return True, "Valid beacon configuration"


# TODO: match values should be returned in the event
def beacon(config):
    """
    Read the log file and return match whole string

    .. code-block:: yaml

        beacons:
          log:
            - file: <path>
            - tags:
                <tag>:
                  regex: <pattern>

    .. note::

        regex matching is based on the `re`_ module

    .. _re: https://docs.python.org/3.6/library/re.html#regular-expression-syntax

    The defined tag is added to the beacon event tag.
    This is not the tag in the log.

    .. code-block:: yaml

        beacons:
          log:
            - file: /var/log/messages #path to log.
            - tags:
                goodbye/world: # tag added to beacon event tag.
                  regex: .*good-bye.* # match good-bye string anywhere in the log entry.
    """
    config = salt.utils.beacons.list_to_dict(config)

    ret = []

    if "file" not in config:
        event = SKEL.copy()
        event["tag"] = "global"
        event["error"] = "file not defined in config"
        ret.append(event)
        return ret

    with salt.utils.files.fopen(config["file"], "r") as fp_:
        loc = __context__.get(LOC_KEY, 0)
        if loc == 0:
            fp_.seek(0, 2)
            __context__[LOC_KEY] = fp_.tell()
            return ret

        fp_.seek(0, 2)
        __context__[LOC_KEY] = fp_.tell()
        fp_.seek(loc)

        txt = fp_.read()
        log.info("txt %s", txt)

        d = {}
        for tag in config.get("tags", {}):
            if "regex" not in config["tags"][tag]:
                continue
            if not config["tags"][tag]["regex"]:
                continue
            try:
                d[tag] = re.compile(r"{}".format(config["tags"][tag]["regex"]))
            except Exception as e:  # pylint: disable=broad-except
                event = SKEL.copy()
                event["tag"] = tag
                event["error"] = "bad regex"
                ret.append(event)

        for line in txt.splitlines():
            for tag, reg in d.items():
                try:
                    m = reg.match(line)
                    if m:
                        event = SKEL.copy()
                        event["tag"] = tag
                        event["raw"] = line
                        event["match"] = "yes"
                        ret.append(event)
                except Exception:  # pylint: disable=broad-except
                    event = SKEL.copy()
                    event["tag"] = tag
                    event["error"] = "bad match"
                    ret.append(event)
    return ret



"""
Manage SQS Queues

.. versionadded:: 2014.7.0

Create and destroy SQS queues. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses ``boto``, which can be installed via package, or pip.

This module accepts explicit SQS credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    sqs.keyid: GKTADJGHEIQSXMKKRBJ08H
    sqs.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile, either
passed in as a dict, or as a string to pull from pillars or minion config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        region: us-east-1

.. code-block:: yaml

    myqueue:
        boto_sqs.present:
            - region: us-east-1
            - keyid: GKTADJGHEIQSXMKKRBJ08H
            - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            - attributes:
                ReceiveMessageWaitTimeSeconds: 20

    # Using a profile from pillars
    myqueue:
        boto_sqs.present:
            - region: us-east-1
            - profile: mysqsprofile

    # Passing in a profile
    myqueue:
        boto_sqs.present:
            - region: us-east-1
            - profile:
                keyid: GKTADJGHEIQSXMKKRBJ08H
                key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
"""

import difflib
import logging

import salt.utils.json

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto is available.
    """
    if "boto_sqs.exists" in __salt__:
        return "boto_sqs"
    return (False, "boto_sqs module could not be loaded")


def present(
    name,
    attributes=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure the SQS queue exists.

    name
        Name of the SQS queue.

    attributes
        A dict of key/value SQS attributes.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": [],
        "changes": {},
    }

    r = __salt__["boto_sqs.exists"](
        name,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if "error" in r:
        ret["result"] = False
        ret["comment"].append(r["error"])
        return ret

    if r["result"]:
        ret["comment"].append("SQS queue {} present.".format(name))
    else:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"].append(
                "SQS queue {} is set to be created.".format(name),
            )
            ret["changes"] = {"old": None, "new": name}
            return ret

        r = __salt__["boto_sqs.create"](
            name,
            attributes=attributes,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        if "error" in r:
            ret["result"] = False
            ret["comment"].append(
                "Failed to create SQS queue {}: {}".format(name, r["error"]),
            )
            return ret

        ret["comment"].append("SQS queue {} created.".format(name))
        ret["changes"]["old"] = None
        ret["changes"]["new"] = name
        # Return immediately, as the create call also set all attributes
        return ret

    if not attributes:
        return ret

    r = __salt__["boto_sqs.get_attributes"](
        name,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if "error" in r:
        ret["result"] = False
        ret["comment"].append(
            "Failed to get queue attributes: {}".format(r["error"]),
        )
        return ret
    current_attributes = r["result"]

    attrs_to_set = {}
    for attr, val in attributes.items():
        _val = current_attributes.get(attr, None)
        if attr == "Policy":
            # Normalize by brute force
            if isinstance(_val, str):
                _val = salt.utils.json.loads(_val)
            if isinstance(val, str):
                val = salt.utils.json.loads(val)
            if _val != val:
                log.debug("Policies differ:\n%s\n%s", _val, val)
                attrs_to_set[attr] = salt.utils.json.dumps(val, sort_keys=True)
        elif str(_val) != str(val):
            log.debug("Attributes differ:\n%s\n%s", _val, val)
            attrs_to_set[attr] = val
    attr_names = ", ".join(attrs_to_set)

    if not attrs_to_set:
        ret["comment"].append("Queue attributes already set correctly.")
        return ret

    final_attributes = current_attributes.copy()
    final_attributes.update(attrs_to_set)

    def _yaml_safe_dump(attrs):
        """
        Safely dump YAML using a readable flow style
        """
        dumper = __utils__["yaml.get_dumper"]("IndentedSafeOrderedDumper")
        return __utils__["yaml.dump"](attrs, default_flow_style=False, Dumper=dumper)

    attributes_diff = "".join(
        difflib.unified_diff(
            _yaml_safe_dump(current_attributes).splitlines(True),
            _yaml_safe_dump(final_attributes).splitlines(True),
        )
    )

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"].append(
            "Attribute(s) {} set to be updated:\n{}".format(
                attr_names,
                attributes_diff,
            )
        )
        ret["changes"] = {"attributes": {"diff": attributes_diff}}
        return ret

    r = __salt__["boto_sqs.set_attributes"](
        name,
        attrs_to_set,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if "error" in r:
        ret["result"] = False
        ret["comment"].append(
            "Failed to set queue attributes: {}".format(r["error"]),
        )
        return ret

    ret["comment"].append(
        "Updated SQS queue attribute(s) {}.".format(attr_names),
    )
    ret["changes"]["attributes"] = {"diff": attributes_diff}
    return ret


def absent(
    name,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure the named sqs queue is deleted.

    name
        Name of the SQS queue.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    r = __salt__["boto_sqs.exists"](
        name,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if "error" in r:
        ret["result"] = False
        ret["comment"] = str(r["error"])
        return ret

    if not r["result"]:
        ret["comment"] = "SQS queue {} does not exist in {}.".format(
            name,
            region,
        )
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "SQS queue {} is set to be removed.".format(name)
        ret["changes"] = {"old": name, "new": None}
        return ret

    r = __salt__["boto_sqs.delete"](
        name,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if "error" in r:
        ret["result"] = False
        ret["comment"] = str(r["error"])
        return ret

    ret["comment"] = "SQS queue {} was deleted.".format(name)
    ret["changes"]["old"] = name
    ret["changes"]["new"] = None
    return ret



import logging
import time

import pytest
from pytestshellutils.exceptions import FactoryNotStarted

log = logging.getLogger(__name__)

ECHO_STR = "The FitnessGram Pacer Test is a multistage aerobic capacity test"

pytestmark = [
    pytest.mark.core_test,
]


def test_basic_command_return(
    salt_mm_minion_1,
    salt_mm_minion_2,
    mm_master_1_salt_cli,
    mm_master_2_salt_cli,
    run_salt_cmds,
):
    """
    Make sure minions return to both masters
    """
    returns = run_salt_cmds(
        [mm_master_1_salt_cli, mm_master_2_salt_cli],
        [salt_mm_minion_1, salt_mm_minion_2],
    )

    assert len(returns) == 4
    assert (mm_master_1_salt_cli, salt_mm_minion_1) in returns
    assert (mm_master_2_salt_cli, salt_mm_minion_1) in returns
    assert (mm_master_1_salt_cli, salt_mm_minion_2) in returns
    assert (mm_master_2_salt_cli, salt_mm_minion_2) in returns


def test_stopped_first_master(
    salt_mm_master_1,
    salt_mm_minion_1,
    salt_mm_minion_2,
    mm_master_1_salt_cli,
    mm_master_2_salt_cli,
    run_salt_cmds,
):
    """
    Make sure minions return only to the second master when the first is stopped
    """
    with salt_mm_master_1.stopped():
        returns = run_salt_cmds(
            [mm_master_1_salt_cli, mm_master_2_salt_cli],
            [salt_mm_minion_1, salt_mm_minion_2],
        )

        assert len(returns) == 2
        assert (mm_master_2_salt_cli, salt_mm_minion_1) in returns
        assert (mm_master_2_salt_cli, salt_mm_minion_2) in returns


def test_stopped_second_master(
    salt_mm_master_2,
    salt_mm_minion_1,
    salt_mm_minion_2,
    mm_master_1_salt_cli,
    mm_master_2_salt_cli,
    run_salt_cmds,
):
    """
    Make sure minions return only to the first master when the second is stopped
    """
    with salt_mm_master_2.stopped():
        returns = run_salt_cmds(
            [mm_master_1_salt_cli, mm_master_2_salt_cli],
            [salt_mm_minion_1, salt_mm_minion_2],
        )

        assert len(returns) == 2
        assert (mm_master_1_salt_cli, salt_mm_minion_1) in returns
        assert (mm_master_1_salt_cli, salt_mm_minion_2) in returns


def test_minion_reconnection_attempts(
    event_listener,
    salt_mm_master_1,
    salt_mm_master_2,
    salt_mm_minion_1,
    salt_mm_minion_2,
    mm_master_1_salt_cli,
    mm_master_2_salt_cli,
    ensure_connections,
):
    """
    Test that minions stay alive and reauth when masters go down and back up, even after restart
    """
    with salt_mm_master_2.stopped():
        with salt_mm_master_1.stopped():
            # Force the minion to restart
            salt_mm_minion_1.terminate()
            # This should make sure the minion stays alive with no masters
            with pytest.raises(FactoryNotStarted):
                with salt_mm_minion_1.started(start_timeout=30):
                    pass

        assert not salt_mm_minion_1.is_running()

        start_time = time.time()
        salt_mm_minion_1.start()

        assert salt_mm_minion_1.is_running()
        assert salt_mm_minion_2.is_running()

        start_events = event_listener.wait_for_events(
            [(salt_mm_master_1.id, "salt/minion/{}/start".format(salt_mm_minion_1.id))],
            timeout=60,
            after_time=start_time,
        )

        assert not start_events.missed
        assert len(start_events.matches) == 1

        ensure_connections([mm_master_1_salt_cli], [salt_mm_minion_1, salt_mm_minion_2])

        start_time = time.time()

    start_events = event_listener.wait_for_events(
        [(salt_mm_master_2.id, "salt/minion/{}/start".format(salt_mm_minion_1.id))],
        timeout=60,
        after_time=start_time,
    )
    assert not start_events.missed
    assert len(start_events.matches) == 1

    ensure_connections(
        [mm_master_1_salt_cli, mm_master_2_salt_cli],
        [salt_mm_minion_1, salt_mm_minion_2],
    )



"""
State module for creating shortcuts on Windows. Handles file shortcuts (`.lnk`)
and url shortcuts (`.url`). Allows for the configuration of icons and hot keys
on file shortcuts. Changing the icon and hot keys are unsupported for url
shortcuts.

.. versionadded:: 3005
"""
import salt.utils.data
import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandExecutionError

# Define the module's virtual name
__virtualname__ = "shortcut"


def __virtual__():
    """
    Only works on Windows systems
    """
    if not salt.utils.platform.is_windows():
        return False, "Shortcut state only available on Windows systems."
    if not __salt__.get("shortcut.create", None):
        return False, "Shortcut state requires the shortcut module."

    return __virtualname__


def present(
    name,
    arguments="",
    description="",
    hot_key="",
    icon_location="",
    icon_index=0,
    target="",
    window_style="Normal",
    working_dir="",
    backup=False,
    force=False,
    make_dirs=False,
    user=None,
):
    r"""
    Create a new shortcut. This can be a file shortcut (``.lnk``) or a url
    shortcut (``.url``).

    Args:

        name (str): The full path to the shortcut

        target (str): The full path to the target

        arguments (str, optional): Any arguments to be passed to the target

        description (str, optional): The description for the shortcut. This is
            shown in the ``Comment`` field of the dialog box. Default is an
            empty string

        hot_key (str, optional): A combination of hot Keys to trigger this
            shortcut. This is something like ``Ctrl+Alt+D``. This is shown in
            the ``Shortcut key`` field in the dialog box. Default is an empty
            string. Available options are:

            - Ctrl
            - Alt
            - Shift
            - Ext

        icon_index (int, optional): The index for the icon to use in files that
            contain multiple icons. Default is 0

        icon_location (str, optional): The full path to a file containing icons.
            This is shown in the ``Change Icon`` dialog box by clicking the
            ``Change Icon`` button. If no file is specified and a binary is
            passed as the target, Windows will attempt to get the icon from the
            binary file. Default is an empty string

        window_style (str, optional): The window style the program should start
            in. This is shown in the ``Run`` field of the dialog box. Default is
            ``Normal``. Valid options are:

            - Normal
            - Minimized
            - Maximized

        working_dir (str, optional): The full path to the working directory for
            the program to run in. This is shown in the ``Start in`` field of
            the dialog box.

        backup (bool, optional): If there is already a shortcut with the same
            name, set this value to ``True`` to backup the existing shortcut and
            continue creating the new shortcut. Default is ``False``

        force (bool, optional): If there is already a shortcut with the same
            name and you aren't backing up the shortcut, set this value to
            ``True`` to remove the existing shortcut and create a new with these
            settings. Default is ``False``

        make_dirs (bool, optional): If the parent directory structure does not
            exist for the new shortcut, create it. Default is ``False``

        user (str, optional): The user to be the owner of any directories
            created by setting ``make_dirs`` to ``True``. If no value is passed
            Salt will use the user account that it is running under. Default is
            an empty string.

    Returns:
        dict: A dictionary containing the changes, comments, and result of the
            state

    Example:

    .. code-block:: yaml

        KB123456:
          wusa.installed:
            - source: salt://kb123456.msu

        # Create a shortcut and set the ``Shortcut key`` (``hot_key``)
        new_shortcut:
          shortcut.present:
            - name: C:\path\to\shortcut.lnk
            - target: C:\Windows\notepad.exe
            - hot_key: Ctrl+Alt+N

        # Create a shortcut and change the icon to the 3rd one in the icon file
        new_shortcut:
          shortcut.present:
            - name: C:\path\to\shortcut.lnk
            - target: C:\Windows\notepad.exe
            - icon_location: C:\path\to\icon.ico
            - icon_index: 2

        # Create a shortcut and change the startup mode to full screen
        new_shortcut:
          shortcut.present:
            - name: C:\path\to\shortcut.lnk
            - target: C:\Windows\notepad.exe
            - window_style: Maximized

        # Create a shortcut and change the icon
        new_shortcut:
          shortcut.present:
            - name: C:\path\to\shortcut.lnk
            - target: C:\Windows\notepad.exe
            - icon_location: C:\path\to\icon.ico

        # Create a shortcut and force it to overwrite an existing shortcut
        new_shortcut:
          shortcut.present:
            - name: C:\path\to\shortcut.lnk
            - target: C:\Windows\notepad.exe
            - force: True

        # Create a shortcut and create any parent directories if they are missing
        new_shortcut:
          shortcut.present:
            - name: C:\path\to\shortcut.lnk
            - target: C:\Windows\notepad.exe
            - make_dirs: True
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": []}

    proposed = {
        "arguments": arguments,
        "description": description,
        "hot_key": hot_key,
        "icon_location": salt.utils.path.expand(icon_location),
        "icon_index": icon_index,
        "path": salt.utils.path.expand(name),
        "target": salt.utils.path.expand(target),
        "window_style": window_style,
        "working_dir": salt.utils.path.expand(working_dir),
    }

    try:
        old = __salt__["shortcut.get"](name)
        changes = salt.utils.data.compare_dicts(old, proposed)
        if not changes:
            ret["comment"] = "Shortcut already present and configured"
            return ret

    except CommandExecutionError:
        changes = {}

    if __opts__["test"]:
        if changes:
            ret["comment"] = "Shortcut will be modified: {}".format(name)
            ret["changes"] = changes
        else:
            ret["comment"] = "Shortcut will be created: {}".format(name)

        ret["result"] = None
        return ret

    try:
        __salt__["shortcut.create"](
            arguments=arguments,
            description=description,
            hot_key=hot_key,
            icon_location=icon_location,
            icon_index=icon_index,
            path=name,
            target=target,
            window_style=window_style,
            working_dir=working_dir,
            backup=backup,
            force=force,
            make_dirs=make_dirs,
            user=user,
        )
    except CommandExecutionError as exc:
        ret["comment"] = ["Failed to create the shortcut: {}".format(name)]
        ret["comment"].append(exc.message)
        ret["result"] = False
        return ret

    try:
        new = __salt__["shortcut.get"](name)
    except CommandExecutionError as exc:
        ret["comment"] = ["Failed to create the shortcut: {}".format(name)]
        ret["comment"].append(exc.message)
        ret["result"] = False
        return ret

    verify_changes = salt.utils.data.compare_dicts(new, proposed)
    if verify_changes:
        ret["comment"] = "Failed to make the following changes:"
        ret["changes"]["failed"] = verify_changes
        ret["result"] = False
        return ret

    if changes:
        ret["comment"] = "Shortcut modified: {}".format(name)
        ret["changes"] = changes
    else:
        ret["comment"] = "Shortcut created: {}".format(name)

    return ret
